
"""
Unit tests for the L{xmantissa.sharing} module.
"""

from zope.interface import Interface, implements

from epsilon.hotfix import require

require("twisted", "trial_assertwarns")

from twisted.trial import unittest

from twisted.python.components import registerAdapter

from twisted.protocols.amp import Command, Box, parseString

from axiom.store import Store
from axiom.item import Item
from axiom.attributes import integer, boolean
from axiom.test.util import QueryCounter
from axiom.userbase import LoginMethod, LoginAccount

from xmantissa import sharing

class IPrivateThing(Interface):
    def mutateSomeState():
        pass

class IExternal(Interface):
    """
    This is an interface for functionality defined externally to an item.
    """
    def doExternal():
        """
        Do something external.
        """


class IExtraExternal(IExternal):
    """
    This is an interface for functionality defined externally to an item.
    """

    def doExternalExtra():
        """
        Do something _extra_ and external.
        """


class IConflicting(Interface):
    """
    This is an interface with a name that conflicts with IExternal.
    """
    def doExternal():
        """
        Conflicts with IExternal.doExternal.
        """



class IReadOnly(Interface):
    def retrieveSomeState():
        """
        Retrieve the data.
        """


class PrivateThing(Item):
    implements(IPrivateThing, IReadOnly)
    publicData = integer(indexed=True)
    externalized = boolean()
    typeName = 'test_sharing_private_thing'
    schemaVersion = 1

    def mutateSomeState(self):
        self.publicData += 5

    def retrieveSomeState(self):
        """
        Retrieve my public data.
        """
        return self.publicData


class ExtraPrivateThing(Item):
    """
    Private class which supports extra operations / adaptations.
    """
    privateData = integer(indexed=True)
    def doPrivateThing(self):
        """
        A method, that does a thing.
        """



class ExternalThing(object):
    implements(IExternal)

    def __init__(self, priv):
        self.privateThing = priv

    def doExternal(self):
        """
        Do an external thing.
        """
        self.privateThing.externalized = True
        return "external"



class ExtraExternalThing(object):
    """
    Adapter for ExtraPrivateThing.
    """
    implements(IExtraExternal)

    def __init__(self, extraPrivateThing):
        self.extraPrivateThing = extraPrivateThing

    # WARNING WARNING WARNING WARNING WARNING

    # The following two methods are shared methods on a shared interfaces which
    # return 'self' as one of the parts of their return value.  IN NORMAL
    # APPLICATION CODE THIS IS A SECURITY RISK!  One of the major features of
    # the sharing system is that it makes programs secure in the face of errors
    # such as accessing an attribute that you are not technically supposed to
    # access from the view.  However, if you give back 'self' to some view code
    # that is asking externally, you are telling the view that it has FULL
    # ACCESS to this object and may call methods on it with impunity.  DO NOT
    # DO THIS IN NORMAL APPLICATION CODE, IT IS ONLY FOR TESTING!!!

    def doExternal(self):
        """
        Do something external and return a 2-tuple of a marker value to indicate
        that this adapter class was used and an identifier to confirm the
        adapter's identity.
        """
        return ("external", self)


    def doExternalExtra(self):
        """
        Similar to doExternal, but a different method
        """
        return ("external-extra", self)

    # DANGER DANGER DANGER DANGER DANGER



registerAdapter(ExternalThing, PrivateThing, IExternal)
registerAdapter(ExtraExternalThing, ExtraPrivateThing, IExtraExternal)


class IPublicThing(Interface):
    def callMethod():
        pass

    def isMethodAvailable(self):
        pass

class PublicFacingAdapter(object):

    implements(IPublicThing)

    def __init__(self, thunk):
        self.thunk = thunk

    def isMethodAvailable(self):
        return IPrivateThing.providedBy(self.thunk)

    def callMethod(self):
        return self.thunk.mutateSomeState()

registerAdapter(PublicFacingAdapter, IReadOnly, IPublicThing)


class InterfaceUtils(unittest.TestCase):
    """
    Test cases for private zope interface utility functionality.
    """
    class A(Interface):
        "A root interface."
    class B(A):
        "A->B"
    class C(B):
        "A->B->C"
    class D(C):
        "A->B->C->D"

    class Q(Interface):
        "Another root interface, unrelated to A"

    class W(B):
        """
        A->B->...
           |
           +->W
        """
    class X(W):
        "A->B->W->X"
    class Y(X):
        "A->B->W->X->Y"
    class Z(Y):
        "A->B->W->X->Y->Z"

    def test_commonParent(self):
        """
        Verify the function which determines the common parent of two interfaces.
        """
        self.assertEquals(sharing._commonParent(self.Z, self.D), self.B)


    def test_noCommonParent(self):
        """
        _commonParent should return None for two classes which do not have a common
        parent.
        """
        self.assertIdentical(sharing._commonParent(self.A, self.Q), None)


    def test_commonParentOfYourself(self):
        """
        The common parent of the same object is itself.
        """
        self.assertIdentical(sharing._commonParent(self.A, self.A), self.A)


class SimpleSharing(unittest.TestCase):


    def setUp(self):
        self.store = Store()


    def test_differentUserSameID(self):
        """
        Verify that if different facets of the same item are shared to different
        users with the same shareID, each user will receive the correct
        respective facet with only the correct methods exposed.
        """
        t = PrivateThing(store=self.store, publicData=789)
        toBob = sharing.shareItem(t, toName=u'bob@example.com',
                                  interfaces=[IReadOnly])
        toAlice = sharing.shareItem(t, toName=u'alice@example.com',
                                    shareID=toBob.shareID,
                                    interfaces=[IPrivateThing])
        # Sanity check.
        self.assertEquals(toBob.shareID, toAlice.shareID)
        asBob = sharing.getShare(self.store,
                                 sharing.getPrimaryRole(
                self.store, u'bob@example.com'),
                                 toBob.shareID)
        asAlice = sharing.getShare(self.store,
                                 sharing.getPrimaryRole(
                self.store, u'alice@example.com'),
                                 toBob.shareID)
        self.assertEquals(asBob.retrieveSomeState(), 789)
        self.assertRaises(AttributeError, lambda : asBob.mutateSomeState)
        self.assertRaises(AttributeError, lambda : asAlice.retrieveSomeState)
        asAlice.mutateSomeState()
        # Make sure they're both seeing the same item.
        self.assertEquals(asBob.retrieveSomeState(), 789+5)


    def test_simpleShareMethods(self):
        """
        Verify that an item which is shared with Role.shareItem can be
        retrieved and manipulated with Role.getShare.  This is the new-style
        API, which isn't yet widely used, but should be preferred in new code.
        """
        t = PrivateThing(store=self.store, publicData=456)
        bob = sharing.getPrimaryRole(self.store, u'bob@example.com',
                                     createIfNotFound=True)
        shareItemResult = bob.shareItem(t)
        gotShare = bob.getShare(shareItemResult.shareID)
        gotShare.mutateSomeState()
        self.assertEquals(t.publicData, 456 + 5)


    def test_simpleShare(self):
        """
        Verify that an item which is shared with shareItem can be retrieved and
        manipulated with getShare.  This is an older-style API, on its way to
        deprecation.
        """
        t = PrivateThing(store=self.store, publicData=456)
        shareItemResult = self.assertWarns(
            PendingDeprecationWarning,
            "Use Role.shareItem() instead of sharing.shareItem().",
            __file__,
            lambda : sharing.shareItem(t, toName=u'bob@example.com'))
        bob = sharing.getPrimaryRole(self.store, u'bob@example.com')
        gotShare = self.assertWarns(
            PendingDeprecationWarning,
            "Use Role.getShare() instead of sharing.getShare().",
            __file__,
            lambda :
                sharing.getShare(self.store, bob, shareItemResult.shareID))
        gotShare.mutateSomeState()
        self.assertEquals(t.publicData, 456 + 5)


    def test_invalidShareID(self):
        """
        Verify that NoSuchShare is raised when getShare is called without sharing
        anything first.
        """
        self.assertRaises(sharing.NoSuchShare,
                          sharing.getShare,
                          self.store,
                          sharing.getPrimaryRole(self.store,
                                                 u'nobody@example.com'),
                          u"not a valid shareID")

    def test_unauthorizedAccessNoShare(self):
        """
        Verify that NoSuchShare is raised when getShare is called with a user who
        is not allowed to access a shared item.
        """
        t = PrivateThing(store=self.store, publicData=345)
        theShare = sharing.shareItem(t, toName=u'somebody@example.com')
        self.assertRaises(sharing.NoSuchShare,
                          sharing.getShare,
                          self.store,
                          sharing.getPrimaryRole(self.store,
                                                 u'nobody@example.com'),
                          theShare.shareID)


    def test_deletedOriginalNoShare(self):
        """
        NoSuchShare should be raised when getShare is called with an item who is
        not allowed to access a shared item.
        """
        t = PrivateThing(store=self.store, publicData=234)
        theShare = sharing.shareItem(t, toName=u'somebody@example.com')
        t.deleteFromStore()
        self.assertRaises(sharing.NoSuchShare,
                          sharing.getShare,
                          self.store,
                          sharing.getPrimaryRole(self.store,
                                                 u'somebody@example.com'),
                          theShare.shareID)


    def test_shareAndAdapt(self):
        """
        Verify that when an item is shared to a particular user with a particular
        interface, retrieving it for that user results in methods on the given
        interface being callable and other methods being restricted.
        """
        t = PrivateThing(store=self.store, publicData=789)

        # Sanity check.
        self.failUnless(IPublicThing(t).isMethodAvailable())

        shared = sharing.shareItem(t, toName=u'testshare', interfaces=[IReadOnly])
        proxy = sharing.getShare(self.store,
                                 sharing.getPrimaryRole(self.store, u'testshare'),
                                 shared.shareID)
        self.assertFalse(IPublicThing(proxy).isMethodAvailable())
        self.assertRaises(AttributeError, IPublicThing(proxy).callMethod)


    def test_getShareProxyWithAdapter(self):
        """
        When you share an item with an interface that has an adapter for that
        interface, the object that results from getShare should provide the
        interface by exposing the adapter rather than the original item.
        """
        privateThing = PrivateThing(store=self.store)
        shared = sharing.shareItem(privateThing, toName=u'testshare',
                                   interfaces=[IExternal])
        proxy = sharing.getShare(self.store,
                                 sharing.getPrimaryRole(self.store, u'testshare'),
                                 shared.shareID)
        proxy.doExternal()
        self.assertTrue(privateThing.externalized)


    def test_conflictingNamesException(self):
        """
        When you share an item with two interfaces that contain different
        adapters with conflicting names, an exception should be raised alerting
        you to this conflict.
        """
        extraPrivateThing = ExtraPrivateThing(store=self.store)
        self.assertRaises(sharing.ConflictingNames,
                          sharing.shareItem, extraPrivateThing, toName=u'testshare',
                          interfaces=[IExternal, IConflicting])


    def test_coalesceInheritedAdapters(self):
        """
        If multiple interfaces that are part of the same inheritance hierarchy are
        specified, only the leaf interfaces should be adapted to, and provided
        for all interfaces it inherits from.
        """

        extraPrivateThing = ExtraPrivateThing(store=self.store)
        role = sharing.getPrimaryRole(self.store, u'testshare')
        extraProxy = sharing.getShare(
            self.store, role, sharing.shareItem(
                extraPrivateThing,
                toRole=role, interfaces=[IExternal,
                                      IExtraExternal]).shareID)

        externalTag, externalObj = extraProxy.doExternal()
        extraExternalTag, extraExternalObj = extraProxy.doExternalExtra()
        self.assertIdentical(externalObj, extraExternalObj)
        self.assertEquals(externalTag, 'external')
        self.assertEquals(extraExternalTag, 'external-extra')



class AccessibilityQuery(unittest.TestCase):

    def setUp(self):
        self.i = 0
        self.store = Store()
        self.things = []
        self.bobThings = []
        self.aliceThings = []
        self.bob = sharing.getPrimaryRole(self.store, u'bob@example.com',
                                          createIfNotFound=True)
        self.alice = sharing.getPrimaryRole(self.store, u'alice@example.com',
                                            createIfNotFound=True)


    def test_twoInterfacesTwoGroups(self):
        """
        Verify that when an item is shared to two roles that a user is a member of,
        they will have access to both interfaces when it is retrieved with
        getShare.
        """
        self.addSomeThings()
        us = sharing.getPrimaryRole(self.store, u'us', True)
        them = sharing.getPrimaryRole(self.store, u'them', True)
        self.bob.becomeMemberOf(us)
        self.bob.becomeMemberOf(them)
        it = PrivateThing(store=self.store, publicData=1234)
        sharing.shareItem(it, toRole=us, shareID=u'q', interfaces=[IPrivateThing])
        sharing.shareItem(it, toRole=them, shareID=u'q', interfaces=[IReadOnly])
        that = sharing.getShare(self.store, self.bob, u'q')
        self.assertEquals(that.retrieveSomeState(), 1234)
        that.mutateSomeState()
        self.assertEquals(that.retrieveSomeState(), 1239)


    def test_twoInterfacesTwoGroupsQuery(self):
        """
        Verify that when an item is shared to two roles that a user is a member of,
        and then retrieved by an asAccessibleTo query, both interfaces will be
        accessible on each object in the query result, and the same number of
        items will be accessible in the query as were shared.
        """
        us = sharing.getPrimaryRole(self.store, u'us', True)
        them = sharing.getPrimaryRole(self.store, u'them', True)
        self.bob.becomeMemberOf(us)
        self.bob.becomeMemberOf(them)
        for x in range(3):
            it = PrivateThing(store=self.store, publicData=x)
            sharing.shareItem(it, toRole=us, shareID=u'q',
                              interfaces=[IPrivateThing])
            sharing.shareItem(it, toRole=them, shareID=u'q',
                              interfaces=[IReadOnly])
        # sanity check
        self.assertEquals(self.store.query(PrivateThing).count(), 3)
        aat = list(sharing.asAccessibleTo(self.bob, self.store.query(
                    PrivateThing, sort=PrivateThing.publicData.descending)))
        aat2 = list(sharing.asAccessibleTo(self.bob, self.store.query(
                    PrivateThing, sort=PrivateThing.publicData.ascending)))
        # sanity check x2
        for acc in aat:
            acc.mutateSomeState()
        expectedData = [x + 5 for x in reversed(range(3))]
        self.assertEquals([acc.retrieveSomeState() for acc in aat],
                          expectedData)
        self.assertEquals([acc.retrieveSomeState() for acc in aat2],
                          list(reversed(expectedData)))


    def test_twoInterfacesTwoGroupsUnsortedQuery(self):
        """
        Verify that when duplicate shares exist for the same item and an
        asAccessibleTo query is made with no specified sort, the roles are
        still deduplicated properly.
        """
        us = sharing.getPrimaryRole(self.store, u'us', True)
        them = sharing.getPrimaryRole(self.store, u'them', True)
        self.bob.becomeMemberOf(us)
        self.bob.becomeMemberOf(them)
        for x in range(3):
            it = PrivateThing(store=self.store, publicData=x)
            sharing.shareItem(it, toRole=us, shareID=u'q',
                              interfaces=[IPrivateThing])
            sharing.shareItem(it, toRole=them, shareID=u'q',
                              interfaces=[IReadOnly])
        # sanity check
        self.assertEquals(self.store.query(PrivateThing).count(), 3)
        aat = list(sharing.asAccessibleTo(self.bob, self.store.query(
                    PrivateThing)))
        # sanity check x2
        for acc in aat:
            acc.mutateSomeState()
        expectedData = [x + 5 for x in range(3)]
        aat.sort(key=lambda i: i.retrieveSomeState())
        self.assertEquals([acc.retrieveSomeState() for acc in aat],
                          expectedData)


    def addSomeThings(self):
        privateThing = PrivateThing(store=self.store, publicData=-self.i)
        self.i += 1
        self.things.append(privateThing)
        self.bobThings.append(sharing.shareItem(
                privateThing, toName=u'bob@example.com',
                interfaces=[IReadOnly]))
        self.aliceThings.append(sharing.shareItem(
                privateThing,
                toName=u'alice@example.com',
                interfaces=[IPrivateThing]))


    def test_asAccessibleTo(self):
        """
        Ensure that L{Role.asAccessibleTo} returns only items actually
        accessible to the given role.
        """
        for i in range(10):
            self.addSomeThings()

        query = self.store.query(PrivateThing)
        aliceQuery = list(self.alice.asAccessibleTo(query))
        bobQuery = list(self.bob.asAccessibleTo(query))

        self.assertEqual(map(sharing.itemFromProxy, bobQuery),
                         map(lambda x: x.sharedItem, self.bobThings))
        self.assertEqual(map(sharing.itemFromProxy, aliceQuery),
                         map(lambda x: x.sharedItem, self.aliceThings))

        self.assertEqual([p.sharedInterfaces
                          for p in aliceQuery], [[IPrivateThing]] * 10)
        self.assertEqual([p.sharedInterfaces
                          for p in bobQuery], [[IReadOnly]] * 10)




    def test_accessibilityQuery(self):
        """
        Ensure that asAccessibleTo returns only items actually accessible to
        the given role.
        """
        for i in range(10):
            self.addSomeThings()

        query = self.store.query(PrivateThing)
        aliceQuery = self.assertWarns(
            PendingDeprecationWarning,
            "Use Role.asAccessibleTo() instead of sharing.asAccessibleTo().",
            __file__,
            lambda : list(sharing.asAccessibleTo(self.alice, query)))
        bobQuery = list(sharing.asAccessibleTo(self.bob, query))

        self.assertEqual(map(sharing.itemFromProxy, bobQuery),
                         map(lambda x: x.sharedItem, self.bobThings))
        self.assertEqual(map(sharing.itemFromProxy, aliceQuery),
                         map(lambda x: x.sharedItem, self.aliceThings))

        self.assertEqual([p.sharedInterfaces
                          for p in aliceQuery], [[IPrivateThing]] * 10)
        self.assertEqual([p.sharedInterfaces
                          for p in bobQuery], [[IReadOnly]] * 10)


    def test_sortOrdering(self):
        """
        Ensure that asAccessibleTo respects query sort order.
        """
        for i in range(10):
            self.addSomeThings()

        query = self.store.query(PrivateThing,
                                 sort=PrivateThing.publicData.ascending)
        # Sanity check.
        self.assertEquals([x.publicData for x in query], range(-9, 1, 1))
        bobQuery = list(sharing.asAccessibleTo(self.bob, query))
        self.assertEquals([x.retrieveSomeState() for x in bobQuery],
                          range(-9, 1, 1))
        query2 = self.store.query(PrivateThing,
                                  sort=PrivateThing.publicData.descending)
        # Sanity check #2
        self.assertEquals([x.publicData for x in query2], range(-9, 1, 1)[::-1])
        bobQuery2 = list(sharing.asAccessibleTo(self.bob, query2))
        self.assertEquals([x.retrieveSomeState() for x in bobQuery2], range(-9, 1, 1)[::-1])


    def test_limit(self):
        """
        Ensure that asAccessibleTo respects query limits.
        """
        for i in range(10):
            self.addSomeThings()

        query = self.store.query(PrivateThing, limit=3)
        bobQuery = list(sharing.asAccessibleTo(self.bob, query))
        self.assertEquals(len(bobQuery), 3)


    def test_limitGetsAllInterfaces(self):
        """
        asAccessibleTo should always collate interfaces together, regardless of
        its limit parameter.
        """
        t = PrivateThing(store=self.store, publicData=self.i)
        sharing.shareItem(t, toName=u'bob@example.com',
                          interfaces=[IPrivateThing], shareID=u'test')
        sharing.shareItem(t, toName=u'Everyone',
                          interfaces=[IReadOnly], shareID=u'test')
        L = list(sharing.asAccessibleTo(
                self.bob, self.store.query(PrivateThing, limit=1)))
        self.assertEquals(len(L), 1)
        self.assertEquals(set(L[0].sharedInterfaces),
                          set([IReadOnly, IPrivateThing]))


    def test_limitMultiShare(self):
        """
        asAccessibleTo should stop after yielding the limit number of results,
        even if there are more shares examined than results.
        """
        L = []
        for x in range(10):
            t = PrivateThing(store=self.store, publicData=self.i)
            L.append(t)
            self.i += 1
            sharing.shareItem(t, toName=u'bob@example.com',
                              interfaces=[IPrivateThing],
                              shareID=unicode(x))
            sharing.shareItem(t, toName=u'Everyone', interfaces=[IReadOnly],
                              shareID=unicode(x))
        proxies = list(sharing.asAccessibleTo(
                self.bob,
                self.store.query(PrivateThing, limit=5,
                                 sort=PrivateThing.publicData.ascending)))
        self.assertEquals(map(sharing.itemFromProxy, proxies), L[:5])
        for proxy in proxies:
            self.assertEquals(set(proxy.sharedInterfaces),
                              set([IPrivateThing, IReadOnly]))


    def test_limitWithPrivateStuff(self):
        """
        Verify that a limited query with some un-shared items will return up to
        the provided limit number of shared items.
        """
        L = []

        def makeThing(shared):
            t = PrivateThing(store=self.store, publicData=self.i)
            self.i += 1
            if shared:
                sharing.shareItem(
                    t, toRole=self.bob, interfaces=[IPrivateThing],
                    shareID=unicode(self.i))
            L.append(t)
        # 0, 1, 2: shared
        for x in range(3):
            makeThing(True)
        # 3, 4, 5: private
        for x in range(3):
            makeThing(False)
        # 6, 7, 8: shared again
        for x in range(3):
            makeThing(True)

        self.assertEquals(
            map(sharing.itemFromProxy,
                sharing.asAccessibleTo(
                    self.bob, self.store.query(
                        PrivateThing, limit=5))),
            [L[0], L[1], L[2], L[6], L[7]])


    def test_limitEfficiency(self):
        """
        Verify that querying a limited number of shared items does not become
        slower as more items are shared.
        """
        zomg = QueryCounter(self.store)

        for i in range(10):
            self.addSomeThings()

        query = self.store.query(
            PrivateThing, limit=3, sort=PrivateThing.publicData.ascending)
        checkit = lambda : list(sharing.asAccessibleTo(self.bob, query))
        before = zomg.measure(checkit)

        for i in range(10):
            self.addSomeThings()

        after = zomg.measure(checkit)
        self.assertEquals(before, after)
    test_limitEfficiency.todo = (
        'currently gets too many results because we should be using paginate')


class HeuristicTestCases(unittest.TestCase):
    """
    These are tests for sharing APIs which heuristically determine identifying
    information about a user's store or a shared item.
    """

    def setUp(self):
        """
        Set up a store for testing.
        """
        self.store = Store()
        self.account = LoginAccount(store=self.store, password=u'1234')
        self.method = LoginMethod(store=self.store,
                                  account=self.account,
                                  localpart=u'username',
                                  domain=u'domain.example.com',
                                  internal=True,
                                  protocol=u'*',
                                  verified=True)


    def test_getSelfRole(self):
        """
        The self-role of a store should be determined by its L{LoginMethod}s.
        """
        self.assertEquals(list(self.store.query(sharing.Role)), [])
        me = sharing.getSelfRole(self.store)
        self.assertEquals(me.externalID, u'username@domain.example.com')
        self.assertEquals(me.store, self.store)
        self.assertEquals(list(me.allRoles()),
                          [me,
                           sharing.getAuthenticatedRole(self.store),
                           sharing.getEveryoneRole(self.store)])


    def test_getAccountRole(self):
        """
        L{getAccountRole} returns a L{Role} in a given store for one of the
        account names passed to it.
        """
        role = sharing.getAccountRole(
            self.store, [(u"username", u"domain.example.com")])
        self.assertEquals(role.externalID, u"username@domain.example.com")


    def test_noAccountRole(self):
        """
        L{getAccountRole} raises L{ValueError} if passed an empty list of account names.
        """
        self.assertRaises(ValueError, sharing.getAccountRole, self.store, [])


    def test_identifierFromSharedItem(self):
        """
        L{sharing.Identifier.fromSharedItem} should identify a shared Item's shareID.
        """
        t = PrivateThing(store=self.store)
        sharing.getEveryoneRole(self.store).shareItem(t, shareID=u'asdf')
        sid = sharing.Identifier.fromSharedItem(t)
        self.assertEquals(sid.shareID, u'asdf')
        self.assertEquals(sid.localpart, u'username')
        self.assertEquals(sid.domain, u'domain.example.com')


    def test_identifierFromSharedItemMulti(self):
        """
        L{sharing.Identifier.fromSharedItem} should identify a shared Item's
        shareID even if it is shared multiple times.
        """
        t = PrivateThing(store=self.store)
        sharing.getEveryoneRole(self.store).shareItem(t, shareID=u'asdf')
        sharing.getAuthenticatedRole(self.store).shareItem(t, shareID=u'jkl;')
        sid = sharing.Identifier.fromSharedItem(t)
        self.assertIn(sid.shareID, [u'asdf', u'jkl;'])
        self.assertEquals(sid.localpart, u'username')
        self.assertEquals(sid.domain, u'domain.example.com')


    def test_identifierFromSharedItemNoShares(self):
        """
        L{sharing.Identifier.fromSharedItem} should raise L{NoSuchShare} if the given
        item is not shared.
        """
        t = PrivateThing(store=self.store)
        self.assertRaises(sharing.NoSuchShare, sharing.Identifier.fromSharedItem, t)


    def test_identifierFromSharedItemNoMethods(self):
        """
        L{sharing.Identifier.fromSharedItem} should raise L{NoSuchShare} if the given
        item's store contains no L{LoginMethod} objects.
        """
        self.method.deleteFromStore()
        t = PrivateThing(store=self.store)
        sharing.getEveryoneRole(self.store).shareItem(t, shareID=u'asdf')
        self.assertRaises(sharing.NoSuchShare, sharing.Identifier.fromSharedItem, t)



class CommandWithIdentifier(Command):
    """
    This command has an Identifier as one of its arguments.
    """
    arguments = [('shareIdentTest', sharing.IdentifierArgument())]



class IdentifierTestCases(unittest.TestCase):
    """
    Tests for the behavior of L{xmantissa.sharing.Identifier}
    """

    def setUp(self):
        """
        Create a few identifiers.
        """
        self.aliceObject = sharing.Identifier(u'object', u'alice', u'example.com')
        self.mostlyAlice = sharing.Identifier(u'not-the-object',
                                      u'alice', u'example.com')
        self.otherAlice = sharing.Identifier(u'object', u'alice', u'example.com')


    def test_equivalence(self):
        """
        Two L{sharing.Identifier} objects that identify the same shared item should
        compare the same.
        """
        self.assertEquals(self.aliceObject, self.otherAlice)
        self.assertFalse(self.aliceObject != self.otherAlice)


    def test_nonEquivalence(self):
        """
        Two L{sharing.Identifier} objects that identify the same shared item should
        compare as not equal.
        """
        self.assertNotEqual(self.aliceObject, self.mostlyAlice)


    def test_otherTypes(self):
        """
        Other types should not compare equal to an L{sharing.Identifier}.
        """
        self.assertNotEqual(self.aliceObject, object())



class IdentifierArgumentTests(unittest.TestCase):
    """
    Tests for serialization and unserialization of
    L{xmantissa.sharing.Identifier} using
    L{xmantissa.sharing.IdentifierArgument}.
    """

    def setUp(self):
        """
        Set up an identifier and its expected serialized form.
        """
        self.identifier = sharing.Identifier(
            u'\u1234object', u'alice', u'example.com')
        self.expectedData = Box(
            shareIdentTest=Box(shareID=u'\u1234object'.encode('utf-8'),
                               localpart=u'alice'.encode('utf-8'),
                               domain=u'example.com'.encode('utf-8')
                               ).serialize()
            ).serialize()


    def test_parse(self):
        """
        L{sharing.IdentifierArgument} should be able to serialize an
        L{Identifier} as an AMP argument to a box.
        """
        outputBox = CommandWithIdentifier.makeArguments(
            dict(shareIdentTest=self.identifier),
            None)
        outputData = outputBox.serialize()
        # Assert on the intermediate / serialized state to make sure the
        # protocol remains stable.
        self.assertEquals(self.expectedData, outputData)


    def test_unparse(self):
        """
        L{sharing.IdentifierArgument} should be able to unserialize an
        L{Identifier} from a serialized box.
        """
        argDict = CommandWithIdentifier.parseArguments(
            parseString(self.expectedData)[0], None)
        self.assertEquals(argDict, dict(shareIdentTest=self.identifier))





"""
Tests for L{xmantissa.test.rendertools}.
"""

from twisted.trial.unittest import TestCase

from nevow.athena import LiveFragment, LiveElement
from nevow.loaders import stan
from nevow.tags import p, directive

from xmantissa.test.rendertools import renderLiveFragment


class LivePageRendererTestCase(TestCase):
    """
    Test utility function L{render} to make sure it can render various kinds of
    fragments.
    """

    message = 'Hello, world.'

    def docFactory(self, renderer, message):
        return stan(p(render=directive(renderer))[message])

    def testRenderLiveFragment(self):
        """
        Test that L{render} spits out the right thing for a L{LiveFragment}.
        """
        docFactory = self.docFactory('liveFragment', self.message)
        self.assertIn(
            self.message,
            renderLiveFragment(LiveFragment(docFactory=docFactory)))


    def testRenderLiveElement(self):
        """
        Test that L{render} spits out the right thing for a L{LiveElement}.
        """
        docFactory = self.docFactory('liveElement', self.message)
        self.assertIn(
            self.message,
            renderLiveFragment(LiveElement(docFactory=docFactory)))



from zope.interface import implements

from nevow import inevow, athena
from nevow.taglibrary import tabbedPane

from xmantissa import ixmantissa

class PatternDictionary(object):
    def __init__(self, docFactory):
        self.docFactory = inevow.IQ(docFactory)
        self.patterns = dict()

    def __getitem__(self, i):
        if i not in self.patterns:
            self.patterns[i] = self.docFactory.patternGenerator(i)
        return self.patterns[i]

def dictFillSlots(tag, slotmap):
    for (k, v) in slotmap.iteritems():
        tag = tag.fillSlots(k, v)
    return tag

class FragmentCollector(athena.LiveFragment):
    implements(ixmantissa.INavigableFragment)

    fragmentName = None
    live = 'athena'
    title = None

    def __init__(self, translator, docFactory=None, collect=(), name='default'):
        self.name = name
        athena.LiveFragment.__init__(self, None, docFactory)

        tabs = []
        for frag in collect:
            if frag.docFactory is None:
                frag.docFactory = translator.getDocFactory(frag.fragmentName, None)
            tabs.append((frag.title, frag))
        self.tabs = tabs

    def head(self):
        for (tabTitle, fragment) in self.tabs:
            fragment.setFragmentParent(self)
            content = fragment.head()
            if content is not None:
                yield content

        yield tabbedPane.tabbedPaneGlue.inlineCSS

    def render_tabbedPane(self, ctx, data):
        tpf = tabbedPane.TabbedPaneFragment(self.tabs, name=self.name)
        tpf.setFragmentParent(self)
        return tpf




from axiom.test.historic import stubloader
from xmantissa.prefs import DefaultPreferenceCollection

class DefaultPreferenceCollectionTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        pc = self.store.findUnique(DefaultPreferenceCollection)
        self.assertEqual(pc.timezone, 'US/Eastern')




"""
Test for upgrading a WebSite by giving it a hostname attribute.
"""

from xmantissa.test.historic.test_website4to5 import WebSiteUpgradeTests

# Subclass it to make a TestCase with a __module__ which won't confuse trial
# and to make stub discovery work correctly.  These two things are implicitly
# discovered from the class definition are unfortunate. -exarkun
class WebSiteUpgradeTests(WebSiteUpgradeTests):
    # This website was so old, it had no hostname information.  So a default
    # will be filled in. -exarkun
    expectedHostname = u"localhost"



# -*- test-case-name: xmantissa.test.test_admin -*-

import random, string

from zope.interface import implements

from twisted.python.components import registerAdapter
from twisted.python.util import sibpath
from twisted.python.filepath import FilePath
from twisted.python import log
from twisted.application.service import IService, Service
from twisted.conch import manhole
from twisted.cred.portal import IRealm

from nevow.page import renderer
from nevow.athena import expose

from epsilon import extime

from axiom.attributes import (integer, boolean, timestamp, bytes, reference,
    inmemory, AND)
from axiom.item import Item, declareLegacyItem
from axiom import userbase
from axiom.dependency import installOn, dependsOn
from axiom.upgrade import registerUpgrader

from xmantissa.liveform import LiveForm, Parameter, ChoiceParameter
from xmantissa.liveform import TEXT_INPUT ,CHECKBOX_INPUT
from xmantissa import webtheme, liveform, webnav, offering, signup
from xmantissa.port import TCPPort, SSLPort
from xmantissa.product import ProductConfiguration, Product, Installation
from xmantissa.suspension import suspendJustTabProviders, unsuspendTabProviders
from xmantissa.tdb import AttributeColumn
from xmantissa.scrolltable import ScrollingFragment
from xmantissa.webapp import PrivateApplication
from xmantissa.website import WebSite, PrefixURLMixin
from xmantissa.terminal import TerminalManhole
from xmantissa.ixmantissa import (
    INavigableElement, INavigableFragment, ISessionlessSiteRootPlugin,
    IProtocolFactoryFactory)

from nevow import rend, athena, static, tags as T


class DeveloperSite(Item, PrefixURLMixin):
    """
    Provides static content sessionlessly for the developer application.
    """
    implements(ISessionlessSiteRootPlugin)

    typeName = 'developer_site'
    schemaVersion = 1

    sessionless = True

    prefixURL = 'static/webadmin'

    # Counts of each kind of user.  These are not maintained, they should be
    # upgraded away at some point.  -exarkun
    developers = integer(default=0)
    administrators = integer(default=0)

    def createResource(self):
        return static.File(sibpath(__file__, 'static'))



class AdminStatsApplication(Item):
    """
    Obsolete.  Only present for schema compatibility.  Do not use.
    """
    powerupInterfaces = (INavigableElement,)
    implements(INavigableElement)

    schemaVersion = 2
    typeName = 'administrator_application'

    updateInterval = integer(default=5)
    privateApplication = dependsOn(PrivateApplication)

    def getTabs(self):
        return []

declareLegacyItem(AdminStatsApplication, 1,
                  dict(updateInterval=integer(default=5)))

def _adminStatsApplication1to2(old):
    new = old.upgradeVersion(AdminStatsApplication.typeName, 1, 2,
                             updateInterval=old.updateInterval,
                             privateApplication=old.store.findOrCreate(PrivateApplication))
    return new
registerUpgrader(_adminStatsApplication1to2, AdminStatsApplication.typeName, 1, 2)

class LocalUserBrowser(Item):
    """
    XXX I am an unfortunate necessity.

    This class shouldn't exist, and in fact, will be destroyed at the first
    possible moment.  It's stateless, existing only to serve as a web lookup
    hook for the UserInteractionFragment view class.
    """

    implements(INavigableElement)

    typeName = 'local_user_browser'
    schemaVersion = 2

    privateApplication = dependsOn(PrivateApplication)
    powerupInterfaces = (INavigableElement,)

    def getTabs(self):
        return [webnav.Tab('Admin', self.storeID, 0.0,
                           [webnav.Tab('Local Users', self.storeID, 0.1)],
                           authoritative=False)]

declareLegacyItem(LocalUserBrowser.typeName, 1,
                  dict(garbage=integer(default=0)))

def _localUserBrowser1to2(old):
    new = old.upgradeVersion(LocalUserBrowser.typeName, 1, 2,
                             privateApplication=old.store.findOrCreate(PrivateApplication))
    return new
registerUpgrader(_localUserBrowser1to2, LocalUserBrowser.typeName, 1, 2)

class UserInteractionFragment(webtheme.ThemedElement):
    """
    Contains two other user-interface elements which allow existing users to be
    browsed and new users to be created, respectively.
    """
    fragmentName = 'admin-user-interaction'

    def __init__(self, userBrowser):
        """
        @param userBrowser: a LocalUserBrowser instance
        """
        super(UserInteractionFragment, self).__init__()
        self.browser = userBrowser


    def userBrowser(self, request, tag):
        """
        Render a TDB of local users.
        """
        f = LocalUserBrowserFragment(self.browser)
        f.docFactory = webtheme.getLoader(f.fragmentName)
        f.setFragmentParent(self)
        return f
    renderer(userBrowser)


    def userCreate(self, request, tag):
        """
        Render a form for creating new users.
        """
        userCreator = liveform.LiveForm(
            self.createUser,
            [liveform.Parameter(
                    "localpart",
                    liveform.TEXT_INPUT,
                    unicode,
                    "localpart"),
             liveform.Parameter(
                    "domain",
                    liveform.TEXT_INPUT,
                    unicode,
                    "domain"),
             liveform.Parameter(
                    "password",
                    liveform.PASSWORD_INPUT,
                    unicode,
                    "password")])
        userCreator.setFragmentParent(self)
        return userCreator
    renderer(userCreate)


    def createUser(self, localpart, domain, password=None):
        """
        Create a new, blank user account with the given name and domain and, if
        specified, with the given password.

        @type localpart: C{unicode}
        @param localpart: The local portion of the username.  ie, the
        C{'alice'} in C{'alice@example.com'}.

        @type domain: C{unicode}
        @param domain: The domain portion of the username.  ie, the
        C{'example.com'} in C{'alice@example.com'}.

        @type password: C{unicode} or C{None}
        @param password: The password to associate with the new account.  If
        C{None}, generate a new password automatically.
        """
        loginSystem = self.browser.store.parent.findUnique(userbase.LoginSystem)
        if password is None:
            password = u''.join([random.choice(string.ascii_letters + string.digits) for i in xrange(8)])
        loginSystem.addAccount(localpart, domain, password)

registerAdapter(UserInteractionFragment, LocalUserBrowser, INavigableFragment)



class LocalUserBrowserFragment(ScrollingFragment):
    jsClass = u'Mantissa.Admin.LocalUserBrowser'

    def __init__(self, userBrowser):
        ScrollingFragment.__init__(self, userBrowser.store.parent,
                                   userbase.LoginMethod,
                                   userbase.LoginMethod.domain != None,
                                   (userbase.LoginMethod.localpart,
                                    userbase.LoginMethod.domain,
                                    userbase.LoginMethod.verified),
                                   defaultSortColumn=userbase.LoginMethod.domain,
                                   defaultSortAscending=True)

    def linkToItem(self, item):
        # no IWebTranslator.  better ideas?
        # will (localpart, domain, protocol) always be unique?
        return unicode(item.storeID)

    def itemFromLink(self, link):
        return self.store.getItemByID(int(link))

    def doAction(self, loginMethod, actionClass):
        """
        Show the form for the requested action.
        """
        loginAccount = loginMethod.account
        return actionClass(
            self,
            loginMethod.localpart + u'@' + loginMethod.domain,
            loginAccount)

    def action_installOn(self, loginMethod):
        return self.doAction(loginMethod, EndowFragment)

    def action_uninstallFrom(self, loginMethod):
        return self.doAction(loginMethod, DepriveFragment)

    def action_suspend(self, loginMethod):
        return self.doAction(loginMethod, SuspendFragment)

    def action_unsuspend(self, loginMethod):
        return self.doAction(loginMethod, UnsuspendFragment)

class EndowDepriveFragment(webtheme.ThemedElement):
    fragmentName = 'user-detail'

    def __init__(self, fragmentParent, username, loginAccount, which):
        super(EndowDepriveFragment, self).__init__(fragmentParent)
        self.account = loginAccount
        self.which = which
        self.username = username

    def _endow(self, **kw):
        subs = self.account.avatars.open()
        def endowall():
            for product in kw.values():
                if product is not None:
                    getattr(product, self.which)(subs)
        subs.transact(endowall)

    def productForm(self, request, tag):
        """
        Render a L{liveform.LiveForm} -- the main purpose of this fragment --
        which will allow the administrator to endow or deprive existing users
        using Products.
        """

        def makeRemover(i):
            def remover(s3lected):
                if s3lected:
                    return self.products[i]
                return None
            return remover

        f = liveform.LiveForm(
            self._endow,
            [liveform.Parameter(
                    'products' + str(i),
                    liveform.FORM_INPUT,
                    liveform.LiveForm(
                        makeRemover(i),
                        [liveform.Parameter(
                                's3lected',
                                liveform.RADIO_INPUT,
                                bool,
                                repr(p),
                                )],
                        '',
                        ),
                    )
             for (i, p)
             in enumerate(self.products)],
            self.which.capitalize() + u' ' + self.username)
        f.setFragmentParent(self)
        return f
    renderer(productForm)
class EndowFragment(EndowDepriveFragment):
    def __init__(self, fragmentParent, username, loginAccount):
        EndowDepriveFragment.__init__(self, fragmentParent,
                                      username, loginAccount,
                                      'installProductOn')
        allProducts = list(self.account.store.query(Product))
        self.products = [p for p in allProducts
                    if not self.account.avatars.open().findUnique(Installation,
                                                              Installation.types
                                                              == p.types,
                                                              None)]
        self.desc = "Install on"

class DepriveFragment(EndowDepriveFragment):
    def __init__(self, fragmentParent, username, loginAccount):
        EndowDepriveFragment.__init__(self, fragmentParent,
                                      username, loginAccount,
                                      'removeProductFrom')
        allProducts = list(self.account.store.query(Product))
        self.products = [p for p in allProducts
                    if self.account.avatars.open().findUnique(Installation,
                                                              Installation.types
                                                              == p.types,
                                                              None)]
        self.desc = "Remove from"

class SuspendFragment(EndowDepriveFragment):
    def __init__(self, fragmentParent, username, loginAccount):
        self.func = suspendJustTabProviders
        EndowDepriveFragment.__init__(self, fragmentParent,
                                      username, loginAccount,
                                      'suspend')
        allProducts = list(self.account.store.query(Product))
        self.products = [p for p in allProducts
                    if self.account.avatars.open().findUnique(Installation,
                                           AND(Installation.types == p.types,
                                               Installation.suspended == False,
                                               ), [])]
        self.desc = "Suspend"

    def _endow(self, **kw):
        subs = self.account.avatars.open()
        def suspend():
            for product in kw.values():
                if product is not None:
                    i = subs.findUnique(Installation,
                                        Installation.types == product.types,
                                        None)
                    self.func(i)
        subs.transact(suspend)

class UnsuspendFragment(SuspendFragment):
    def __init__(self, fragmentParent, username, loginAccount):
        self.func = unsuspendTabProviders
        EndowDepriveFragment.__init__(self, fragmentParent,
                                      username, loginAccount,
                                      'unsuspend')
        allProducts = list(self.account.store.query(Product))
        self.products = [p for p in allProducts
                    if self.account.avatars.open().findUnique(Installation,
                                           AND(Installation.types == p.types,
                                               Installation.suspended == True),
                                                 None)]
        self.desc = "Unsuspend"



class DeveloperApplication(Item):
    """
    """
    implements(INavigableElement)

    schemaVersion = 2
    typeName = 'developer_application'

    privateApplication = dependsOn(PrivateApplication)
    statementCount = integer(default=0)
    powerupInterfaces = (INavigableElement,)

    def deletedFromStore(self, *a, **kw):
        return super(DeveloperApplication, self).deletedFromStore(*a, **kw)

    # INavigableElement
    def getTabs(self):
        return [webnav.Tab('Admin', self.storeID, 0.0,
                           [webnav.Tab('REPL', self.storeID, 0.0)],
                           authoritative=False)]

declareLegacyItem(DeveloperApplication.typeName, 1,
                  dict(statementCount=integer(default=0)))

def _developerApplication1to2(old):
    new = old.upgradeVersion(DeveloperApplication.typeName, 1, 2,
                             statementCount=old.statementCount,
                             privateApplication=old.store.findOrCreate(PrivateApplication))
    return new
registerUpgrader(_developerApplication1to2, DeveloperApplication.typeName, 1, 2)

class REPL(athena.LiveFragment):
    """
    Provides an interactive Read-Eval-Print loop. On a web page (duh).
    """
    implements(INavigableFragment)

    jsClass = u'Mantissa.InterpreterWidget'

    fragmentName = 'admin-python-repl'
    live = 'athena'

    def __init__(self, *a, **kw):
        rend.Fragment.__init__(self, *a, **kw)
        self.namespace = {'s': self.original.store, 'getStore': self.getStore}
        self.interpreter = manhole.ManholeInterpreter(
            self,
            self.namespace)

    def getStore(self, name, domain):
        """Convenience method for the REPL. I got tired of typing this string every time I logged in."""
        return IRealm(self.original.store.parent).accountByAddress(name, domain).avatars.open()

    def head(self):
        return ()


    def addOutput(self, output, async=False):
        self.callRemote('addOutputLine', unicode(output, 'ascii'))


    def evaluateInputLine(self, inputLine):
        return self.interpreter.push(inputLine)
    expose(evaluateInputLine)



registerAdapter(REPL, DeveloperApplication, INavigableFragment)

class Traceback(Item):
    typeName = 'mantissa_traceback'
    schemaVersion = 1

    when = timestamp()
    traceback = bytes()
    collector = reference()

    def __init__(self, store, collector, failure):
        when = extime.Time()
        traceback = failure.getTraceback()
        super(Traceback, self).__init__(
            store=store,
            traceback=traceback,
            when=when,
            collector=collector)

class TracebackCollector(Item, Service):
    implements(IService)

    typeName = 'mantissa_traceback_collector'
    schemaVersion = 1

    tracebackCount = integer(default=0)

    parent = inmemory()
    running = inmemory()
    name = inmemory()
    powerupInterfaces = (IService,)

    def installed(self):
        self.setServiceParent(self.store)

    def startService(self):
        log.addObserver(self.emit)

    def stopService(self):
        log.removeObserver(self.emit)

    def emit(self, event):
        if event.get('isError') and event.get('failure') is not None:
            f = event['failure']
            def txn():
                self.tracebackCount += 1
                Traceback(store=self.store, collector=self, failure=f)
            self.store.transact(txn)

    def getTracebacks(self):
        """
        Return an iterable of Tracebacks that have been collected.
        """
        return self.store.query(Traceback,
                                Traceback.collector == self)


class TracebackViewer(Item):
    implements(INavigableElement)

    typeName = 'mantissa_tb_viewer'
    schemaVersion = 2

    allowDeletion = boolean(default=False)

    privateApplication = dependsOn(PrivateApplication)
    powerupInterfaces = (INavigableElement,)

    def getTabs(self):
        return [webnav.Tab('Admin', self.storeID, 0.0,
                           [webnav.Tab('Errors', self.storeID, 0.3)],
                           authoritative=False)]

    def _getCollector(self):
        def ifCreate(coll):
            installOn(coll, self.store.parent)
        return self.store.parent.findOrCreate(TracebackCollector, ifCreate)

    # this needs to be moved somewhere else, topPanelContent is no more
    #def topPanelContent(self):
    #    # XXX There should really be a juice protocol for this.
    #    return '%d errors logged' % (self._getCollector().tracebackCount,)

declareLegacyItem(TracebackViewer, 1,
                  dict(allowDeletion=boolean(default=False)))

def _tracebackViewer1to2(old):
    return old.upgradeVersion(TracebackViewer.typeName, 1, 2,
                              allowDeletion=old.allowDeletion,
                              privateApplication=old.store.findOrCreate(PrivateApplication))
registerUpgrader(_tracebackViewer1to2, TracebackViewer.typeName, 1, 2)


class TracebackViewerFragment(rend.Fragment):
    implements(INavigableFragment)

    live = False
    fragmentName = 'traceback-viewer'

    def head(self):
        return ()

    def render_tracebacks(self, ctx, data):
        for tb in self.original._getCollector().getTracebacks():
            yield T.div[T.code[T.pre[tb.traceback]]]

registerAdapter(TracebackViewerFragment, TracebackViewer, INavigableFragment)



class PortConfiguration(Item):
    """
    Marker powerup which allows those on whom it is installed to modify the
    configuration of listening ports in this server.
    """
    implements(INavigableElement)

    powerupInterfaces = (INavigableElement,)

    # Only present because Axiom requires at least one attribute on an Item.
    garbage = integer(default=12345678653)

    def getTabs(self):
        """
        Add this object to the tab navigation so it can display configuration
        information and allow configuration to be modified.
        """
        return [webnav.Tab('Admin', self.storeID, 0.0,
                           [webnav.Tab('Ports', self.storeID, 0.4)],
                           authoritative=False)]


    def createPort(self, portNumber, ssl, certPath, factory, interface=u''):
        """
        Create a new listening port.

        @type portNumber: C{int}
        @param portNumber: Port number on which to listen.

        @type ssl: C{bool}
        @param ssl: Indicates whether this should be an SSL port or not.

        @type certPath: C{str}
        @param ssl: If C{ssl} is true, a path to a certificate file somewhere
        within the site store's files directory.  Ignored otherwise.

        @param factory: L{Item} which provides L{IProtocolFactoryFactory} which
        will be used to get a protocol factory to associate with this port.

        @return: C{None}
        """
        store = self.store.parent
        if ssl:
            port = SSLPort(store=store, portNumber=portNumber,
                           certificatePath=FilePath(certPath), factory=factory,
                           interface=interface)
        else:
            port = TCPPort(store=store, portNumber=portNumber, factory=factory,
                           interface=interface)
        installOn(port, store)



class FactoryColumn(AttributeColumn):
    """
    Display the name of the class of items referred to by a reference
    attribute.
    """
    def extractValue(self, model, item):
        """
        Get the class name of the factory referenced by a port.

        @param model: Either a TabularDataModel or a ScrollableView, depending
        on what this column is part of.

        @param item: A port item instance (as defined by L{xmantissa.port}).

        @rtype: C{unicode}
        @return: The name of the class of the item to which this column's
        attribute refers.
        """
        factory = super(FactoryColumn, self).extractValue(model, item)
        return factory.__class__.__name__.decode('ascii')



class CertificateColumn(AttributeColumn):
    """
    Display a path attribute as a unicode string.
    """
    def extractValue(self, model, item):
        """
        Get the path referenced by this column's attribute.

        @param model: Either a TabularDataModel or a ScrollableView, depending
        on what this column is part of.

        @param item: A port item instance (as defined by L{xmantissa.port}).

        @rtype: C{unicode}
        """
        certPath = super(CertificateColumn, self).extractValue(model, item)
        return certPath.path.decode('utf-8', 'replace')



class PortScrollingFragment(ScrollingFragment):
    """
    A scrolling fragment for TCPPorts and SSLPorts which knows how to link to
    them and how to delete them.

    @ivar userStore: The store of the user viewing these ports.

    @ivar siteStore: The site store, where TCPPorts and SSLPorts are loaded
    from.
    """
    jsClass = u'Mantissa.Admin.PortBrowser'

    def __init__(self, userStore, portType, columns):
        super(PortScrollingFragment, self).__init__(
            userStore.parent,
            portType,
            None,
            columns)
        self.userStore = userStore
        self.siteStore = userStore.parent
        self.webTranslator = self.userStore.findUnique(PrivateApplication)


    def itemFromLink(self, link):
        """
        @type link: C{unicode}
        @param link: A webID to translate into an item.

        @rtype: L{Item}
        @return: The item to which the given link referred.
        """
        return self.siteStore.getItemByID(self.webTranslator.linkFrom(link))


    def action_delete(self, port):
        """
        Delete the given port.
        """
        port.deleteFromStore()



class PortConfigurationFragment(webtheme.ThemedElement):
    """
    Provide the view for L{PortConfiguration}.

    Specifically, three renderers are offered: the first two, L{tcpPorts} and
    L{sslPorts}, add a L{PortScrollingFragment} to their tag as a child; the
    last, L{createPortForm} adds a L{LiveForm} for adding new ports to its tag
    as a child.

    @ivar portConf: The L{PortConfiguration} item.
    @ivar store: The user store.
    """
    implements(INavigableFragment)

    fragmentName = 'port-configuration'


    def __init__(self, portConf):
        super(PortConfigurationFragment, self).__init__()
        self.portConf = portConf
        self.store = portConf.store


    def head(self):
        return ()


    def tcpPorts(self, req, tag):
        """
        Create and return a L{PortScrollingFragment} for the L{TCPPort} items
        in site store.
        """
        f = PortScrollingFragment(
            self.store,
            TCPPort,
            (TCPPort.portNumber,
             TCPPort.interface,
             FactoryColumn(TCPPort.factory)))
        f.setFragmentParent(self)
        f.docFactory = webtheme.getLoader(f.fragmentName)
        return tag[f]
    renderer(tcpPorts)


    def sslPorts(self, req, tag):
        """
        Create and return a L{PortScrollingFragment} for the L{SSLPort} items
        in the site store.
        """
        f = PortScrollingFragment(
            self.store,
            SSLPort,
            (SSLPort.portNumber,
             SSLPort.interface,
             CertificateColumn(SSLPort.certificatePath),
             FactoryColumn(SSLPort.factory)))
        f.setFragmentParent(self)
        f.docFactory = webtheme.getLoader(f.fragmentName)
        return tag[f]
    renderer(sslPorts)


    def createPortForm(self, req, tag):
        """
        Create and return a L{LiveForm} for adding a new L{TCPPort} or
        L{SSLPort} to the site store.
        """
        def port(s):
            n = int(s)
            if n < 0 or n > 65535:
                raise ValueError(s)
            return n

        factories = []
        for f in self.store.parent.powerupsFor(IProtocolFactoryFactory):
            factories.append((f.__class__.__name__.decode('ascii'),
                              f,
                              False))

        f = LiveForm(
            self.portConf.createPort,
            [Parameter('portNumber', TEXT_INPUT, port, 'Port Number',
                       'Integer 0 <= n <= 65535 giving the TCP port to bind.'),

             Parameter('interface', TEXT_INPUT, unicode, 'Interface',
                       'Hostname to bind to, or blank for all interfaces.'),

             Parameter('ssl', CHECKBOX_INPUT, bool, 'SSL',
                       'Select to indicate port should use SSL.'),

             # Text area?  File upload?  What?
             Parameter('certPath', TEXT_INPUT, unicode, 'Certificate Path',
                       'Path to a certificate file on the server, if SSL is to be used.'),

             ChoiceParameter('factory', factories, 'Protocol Factory',
                             'Which pre-existing protocol factory to associate with this port.')])
        f.setFragmentParent(self)
        # f.docFactory = webtheme.getLoader(f.fragmentName)
        return tag[f]
    renderer(createPortForm)

registerAdapter(PortConfigurationFragment, PortConfiguration, INavigableFragment)



class AdministrativeBenefactor(Item):
    typeName = 'mantissa_administrative_benefactor'
    schemaVersion = 1

    endowed = integer(default=0)
    powerupNames = ["xmantissa.webadmin.AdminStatsApplication",
                    "xmantissa.webadmin.DeveloperApplication",
                    "xmantissa.signup.SignupConfiguration",
                    "xmantissa.webadmin.TracebackViewer",
                    "xmantissa.webadmin.BatchManholePowerup",
                    "xmantissa.webadmin.LocalUserBrowser"]


def endowAdminPowerups(userStore):
    powerups = [
            # Install a web site for the individual user as well.
            # This is necessary because although we have a top-level
            # website for everybody, not all users should be allowed
            # to log in through the web (like UNIX's "system users",
            # "nobody", "database", etc.)  Note, however, that there
            # is no port number, because the WebSite's job in this
            # case is to be a web *resource*, not a web *server*.
            WebSite,

            # Now we install the 'private application' plugin for
            # 'admin', on admin's private store, This provides the URL
            # "/private", but only when 'admin' is logged in.  It is a
            # hook to hang other applications on.  (XXX Rename:
            # PrivateApplication should probably be called
            # PrivateAppShell)
            PrivateApplication,

            # This is a plugin *for* the PrivateApplication; it publishes
            # objects via the tab-based navigation (a Python interactive
            # interpreter).
            DeveloperApplication,

            #ProductConfiguration lets admins collect powerups into
            #Products users can sign up for.

            ProductConfiguration,

            # And another one: SignupConfiguration allows the
            # administrator to add signup forms which grant various
            # kinds of account.
            signup.SignupConfiguration,

            # This one lets the administrator view unhandled
            # exceptions which occur in the server.
            TracebackViewer,

            # Allow the administrator to set the ports associated with
            # different network services.
            PortConfiguration,

            # And this one gives the administrator a page listing all
            # users which exist in this site's credentials database.
            LocalUserBrowser,

            # Grant Python REPL access to the server.
            TerminalManhole,
            ]
    for powerup in powerups:
        installOn(powerup(store=userStore), userStore)
    # This is another PrivateApplication plugin.  It allows
    # the administrator to configure the services offered
    # here.
    oc = offering.OfferingConfiguration(store=userStore)
    installOn(oc, userStore)




"""
Generate a stub for the tests for the upgrade from schema version 1 to 2 of the
Person item.
"""

from epsilon.extime import Time

from axiom.test.historic.stubloader import saveStub

from xmantissa.people import Organizer, Person

NAME = u'testuser'
CREATED = Time.fromPOSIXTimestamp(1234567890)

def createDatabase(store):
    """
    Make a L{Person} in the given store.
    """
    organizer = Organizer(store=store)
    person = Person(
        store=store,
        organizer=organizer,
        name=NAME)
    person.created = CREATED


if __name__ == '__main__':
    saveStub(createDatabase, 13023)



from axiom.test.historic.stubloader import saveStub
from xmantissa.people import EmailAddress, Person

def createDatabase(s):
    EmailAddress(store=s,
                 address=u'bob@divmod.com',
                 type=u'default',
                 person=Person(store=s,
                               name=u'Bob'))

if __name__ == '__main__':
    saveStub(createDatabase, 7052)



"""
Database-creation script for testing the version 2 to version 3 upgrader of
L{Mugshot}.
"""
from twisted.python.filepath import FilePath

from axiom.test.historic.stubloader import saveStub
from xmantissa.people import Mugshot, Person


MUGSHOT_TYPE = u'image/png'
MUGSHOT_BODY_PATH_SEGMENTS = ('mugshot',)



def createDatabase(store):
    """
    Make L{Person} and L{Mugshot} items.  Set the C{body} and C{smallerBody}
    attributes of the L{Mugshot} item to point at a copy of
    I{xmantissa/test/resources/square.png} beneath the store's directory.
    """
    atomicImageFile = store.newFile(*MUGSHOT_BODY_PATH_SEGMENTS)
    imageFilePath = FilePath(__file__).parent().parent().child(
        'resources').child('square.png')
    atomicImageFile.write(imageFilePath.getContent())
    atomicImageFile.close()

    Mugshot(store=store,
            person=Person(store=store),
            body=atomicImageFile.finalpath,
            smallerBody=atomicImageFile.finalpath,
            type=MUGSHOT_TYPE)



if __name__ == '__main__':
    saveStub(createDatabase, 13812)



from zope.interface import implements
from zope.interface import classProvides

from twisted.trial.unittest import TestCase
from twisted.python.reflect import qual
from twisted.python.util import sibpath
from twisted.python.filepath import FilePath, InsecurePath

from nevow.athena import LivePage
from nevow.loaders import stan, xmlstr
from nevow.tags import (
    html, head, body, invisible, directive)
from nevow.context import WovenContext
from nevow.testutil import FakeRequest
from nevow.flat import flatten
from nevow.inevow import IRequest

from nevow.athena import LiveFragment

from axiom.item import Item
from axiom.attributes import integer
from axiom.store import Store
from axiom.substore import SubStore
from axiom.dependency import installOn
from axiom.plugins.mantissacmd import Mantissa

from xmantissa.ixmantissa import (
    ITemplateNameResolver, IOfferingTechnician, ISiteURLGenerator)

from xmantissa import webtheme
from xmantissa.webtheme import (
    getInstalledThemes, MantissaTheme, ThemedFragment,
    ThemedElement, ThemedDocumentFactory,
    SiteTemplateResolver, XHTMLDirectoryTheme)

from xmantissa.offering import Offering, installOffering
from xmantissa.plugins.baseoff import baseOffering

from xmantissa.publicweb import PublicAthenaLivePage
from xmantissa.webapp import GenericNavigationAthenaPage, _PageComponents

from xmantissa.test.test_offering import FakeOfferingTechnician
from xmantissa.test.validation import XHTMLDirectoryThemeTestsMixin


class ThemedDocumentFactoryTests(TestCase):
    """
    Tests for the automatic document factory descriptor,
    L{ThemedDocumentFactory}.
    """
    def test_getter(self):
        """
        Retrieving the value of a L{ThemedDocumentFactory} descriptor should
        cause an L{ITemplateNameResolver} to be requested from the supplied
        callable and a loader for the template for the fragment name the
        descriptor was created with to be created and returned.
        """
        _docFactory = object()
        loadAttempts = []
        fragmentName = 'abc'
        class Dummy(object):
            class StubResolver(object):
                classProvides(ITemplateNameResolver)
                def getDocFactory(name):
                    loadAttempts.append(name)
                    return _docFactory
                getDocFactory = staticmethod(getDocFactory)
            docFactory = ThemedDocumentFactory(fragmentName, 'StubResolver')
        self.assertIdentical(Dummy().docFactory, _docFactory)
        self.assertEqual(loadAttempts, [fragmentName])



class FakeTheme:
    """
    Stub theme object for template-loader tests.
    """
    implements(ITemplateNameResolver)
    def __init__(self, name, priority):
        self.name = name
        self.priority = priority

    def getDocFactory(self, n, default):
        """
        Doesn't have to return anything meaningful, just something
        recognizable for assertions.
        """
        return [self.name, n]


class FakeOffering:
    def __init__(self, name, priority):
        self.themes = [FakeTheme(name, priority)]

class WebThemeTestCase(TestCase):
    def _render(self, element):
        """
        Put the given L{IRenderer} provider into an L{athena.LivePage} and
        render it.  Return a Deferred which fires with the request object used
        which is an instance of L{nevow.testutil.FakeRequest}.
        """
        p = LivePage(
            docFactory=stan(
                html[
                    head(render=directive('liveglue')),
                    body[
                        invisible(render=lambda ctx, data: element)]]))
        element.setFragmentParent(p)

        ctx = WovenContext()
        req = FakeRequest()
        ctx.remember(req, IRequest)

        d = p.renderHTTP(ctx)
        def rendered(ign):
            p.action_close(None)
            return req
        d.addCallback(rendered)
        return d


    def test_getAllThemesPrioritization(self):
        """
        Test that the L{xmantissa.webtheme.getAllThemes} function returns
        L{ITemplateNameResolver} providers from the installed
        L{xmantissa.ixmantissa.IOffering} plugins in priority order.
        """
        lastPriority = None
        for theme in webtheme.getAllThemes():
            if lastPriority is None:
                lastPriority = theme.priority
            else:
                self.failIf(
                    theme.priority > lastPriority,
                    "Theme out of order: %r" % (theme,))
                lastPriority = theme.priority


    def test_getInstalledThemes(self):
        """
        Test that only themes which belong to offerings installed on a
        particular store are returned by
        L{xmantissa.webtheme.getInstalledThemes}.
        """
        s = Store()
        self.assertEquals(getInstalledThemes(s), [])

        installOffering(s, baseOffering, {})

        installedThemes = getInstalledThemes(s)
        self.assertEquals(len(installedThemes), 1)
        self.failUnless(isinstance(installedThemes[0], MantissaTheme))


    def _defaultThemedRendering(self, cls):
        class ThemedSubclass(cls):
            pass
        d = self._render(ThemedSubclass())
        def rendered(req):
            self.assertIn(
                qual(ThemedSubclass),
                req.v)
            self.assertIn(
                'specified no <code>fragmentName</code> attribute.',
                req.v)
        d.addCallback(rendered)
        return d


    def test_themedFragmentDefaultRendering(self):
        """
        Test that a ThemedFragment which does not override fragmentName is
        rendered with some debugging tips.
        """
        return self._defaultThemedRendering(ThemedFragment)


    def test_themedElementDefaultRendering(self):
        """
        Test that a ThemedElement which does not override fragmentName is
        rendered with some debugging tips.
        """
        return self._defaultThemedRendering(ThemedElement)



class MantissaThemeTests(XHTMLDirectoryThemeTestsMixin, TestCase):
    """
    Stock L{XHTMLDirectoryTheme} tests applied to L{baseOffering} and its
    theme.
    """
    offering = baseOffering
    theme = offering.themes[0]



CUSTOM_MSG = xmlstr('<div>Athena unsupported here</div>')
BASE_MSG =  file(sibpath(__file__,
                         "../themes/base/athena-unsupported.html")
                 ).read().strip()



class StubThemeProvider(Item):
    """
    Trivial implementation of a theme provider, for testing that custom
    Athena-unsupported pages can be used.
    """
    _attribute = integer(doc="exists to pacify Axiom's hunger for attributes")
    implements(ITemplateNameResolver)
    powerupInterfaces = (ITemplateNameResolver,)
    def getDocFactory(self, name):
        """
        Return the page indicating Athena isn't available, if requested.
        """
        if name == 'athena-unsupported':
            return CUSTOM_MSG



class AthenaUnsupported(TestCase):
    """
    Tests for proper treatment of browsers that don't support Athena.
    """
    def setUp(self):
        self.siteStore = Store(filesdir=self.mktemp())
        Mantissa().installSite(self.siteStore, u"example.com", u"", False)


    def test_publicPage(self):
        """
        Test that L{publicpage.PublicAthenaLivePage} supports themeing of
        Athena's unsupported-browser page.
        """
        stp = StubThemeProvider(store=self.siteStore)
        installOn(stp, self.siteStore)
        p = PublicAthenaLivePage(self.siteStore, None)
        self.assertEqual(p.renderUnsupported(None),
                         flatten(CUSTOM_MSG))


    def test_navPage(self):
        """
        Test that L{webapp.GenericNavigationLivePage} supports theming of
        Athena's unsupported-browser page based on an L{ITemplateNameResolver}
        installed on the viewing user's store.
        """
        subStore = SubStore.createNew(
            self.siteStore, ['athena', 'unsupported']).open()
        stp = StubThemeProvider(store=subStore)
        installOn(stp, subStore)
        p = GenericNavigationAthenaPage(stp,
                                        LiveFragment(),
                                        _PageComponents([], None, None,
                                                        None, None),
                                        None)
        self.assertEqual(p.renderUnsupported(None),
                         flatten(CUSTOM_MSG))



class Loader(TestCase):
    def setUp(self):
        self._getAllThemes = webtheme.getAllThemes
        self.gATcalled = 0
        def fakeGetAllThemes():
            self.gATcalled += 1
            return [FakeTheme('foo', 7),
                    FakeTheme('baz', 2)]
        webtheme._loaderCache.clear()
        webtheme.getAllThemes = fakeGetAllThemes

    def tearDown(self):
        webtheme.getAllThemes = self._getAllThemes

    def test_getLoader(self):
        """
        getLoader should search available themes for the named
        template and return it.
        """
        self.assertEquals(webtheme.getLoader('template'),
                          ['foo', 'template'])

    def test_getLoaderCaching(self):
        """
        getLoader should return identical loaders for equal arguments.
        """
        self.assertIdentical(webtheme.getLoader('template'),
                             webtheme.getLoader('template'))
        self.assertEqual(self.gATcalled, 1)



class TestSiteTemplateResolver(TestCase):
    """
    Tests for L{SiteTemplateResolver}
    """
    def setUp(self):
        """
        Create a L{Store} with a fake L{IOfferingTechnician} powerup which
        allows fine-grained control of template name resolution.
        """
        self.offeringTech = FakeOfferingTechnician()
        self.store = Store()
        self.store.inMemoryPowerUp(self.offeringTech, IOfferingTechnician)
        self.siteResolver = SiteTemplateResolver(self.store)


    def getDocFactoryWithoutCaching(self, templateName):
        """
        Use C{self.siteResolver} to get a loader for the named template,
        flushing the template cache first in order to make the result reflect
        any changes which in offering or theme availability which may have
        happened since the last call.
        """
        webtheme.theThemeCache.emptyCache()
        return self.siteResolver.getDocFactory(templateName)


    def test_getDocFactory(self):
        """
        L{SiteTemplateResolver.getDocFactory} should return only installed
        themes for its store.
        """
        class FakeTheme(object):
            priority = 0

            def getDocFactory(self, templateName, default=None):
                if templateName == 'shell':
                    return object()
                return default

        self.assertIdentical(self.getDocFactoryWithoutCaching('shell'), None)
        self.offeringTech.installOffering(
            Offering(
                u'an offering', None, [], [], [], [], [FakeTheme()]))
        self.assertNotIdentical(self.getDocFactoryWithoutCaching('shell'), None)



class XHTMLDirectoryThemeTests(TestCase):
    """
    Tests for L{XHTMLDirectoryTheme}.
    """
    def setUp(self):
        """
        Set up the store, a temporary test dir and a theme for the tests.
        """
        self.store = Store()
        self.testDir = FilePath(self.mktemp())
        self.testDir.makedirs()
        self.theme = XHTMLDirectoryTheme(
            'testtheme',
            directoryName=self.testDir.path)


    def test_directoryAttribute(self):
        """
        L{XHTMLDirectoryTheme} should have a directory attribute of type
        L{twisted.python.filepath.FilePath}.
        """
        self.assertEqual(self.theme.directory, self.testDir)
        self.assertEqual(self.theme.directory.path, self.theme.directoryName)


    def test_childFragmentsInGetDocFactory(self):
        """
        L{XHTMLDirectoryTheme.getDocFactory} should handle subdirectories
        sanely, without exposing parent directories.
        """
        fragmentName = 'dir/file'
        child = self.testDir.child('dir')
        child.makedirs()
        child.child('file.html').touch()
        resolvedTemplate = self.theme.getDocFactory(fragmentName)
        foundPath = FilePath(resolvedTemplate.template)
        expectedPath = FilePath(
            "%s/%s.html" % (self.theme.directoryName, fragmentName))
        self.assertEqual(foundPath, expectedPath)
        self.assertRaises(InsecurePath, self.theme.getDocFactory,
                          '../insecure/')


    def test_noStylesheetLocation(self):
        """
        L{XHTMLDirectoryTheme.head} returns C{None} if I{stylesheetLocation} is
        C{None}.
        """
        self.assertIdentical(self.theme.head(None, None), None)


    def test_stylesheetLocation(self):
        """
        L{XHTMLDirectoryTheme.head} returns a link tag which gives the location
        of the stylesheet given by I{stylesheetLocation} if there is one.
        """
        siteStore = Store(filesdir=self.mktemp())
        Mantissa().installSite(siteStore, u"example.com", u"", False)
        site = ISiteURLGenerator(siteStore)

        self.theme.stylesheetLocation = ['foo', 'bar']
        request = FakeRequest()
        link = self.theme.head(request, site)
        self.assertEqual(link.tagName, 'link')
        self.assertEqual(link.attributes['rel'], 'stylesheet')
        self.assertEqual(link.attributes['type'], 'text/css')
        self.assertEqual(
            site.rootURL(request).child('foo').child('bar'),
            link.attributes['href'])



class TestThemeCache(TestCase):
    """
    some tests for L{ThemeCache}.
    """
    def setUp(self):
        """
        Replace L{getOfferings} with a mock method returning some fake
        offerings.
        """
        self._getOfferings = webtheme.getOfferings
        self.called = 0
        def fakeGetOfferings():
            self.called += 1
            return [FakeOffering('foo', 7),
                    FakeOffering('baz', 2),
                    FakeOffering('boz', 5)]

        webtheme.getOfferings = fakeGetOfferings

    def tearDown(self):
        """
        Reset L{getOfferings} to its original value.
        """
        webtheme.getOfferings = self._getOfferings


    def test_getAllThemes(self):
        """
        C{getAllThemes} should collect themes from available
        offerings, and only call C{getOfferings} once no matter how
        many times it's called.
        """
        tc = webtheme.ThemeCache()
        ths = tc.getAllThemes()
        self.assertEqual([theme.name for theme in ths],
                         ['foo', 'boz', 'baz'])
        tc.getAllThemes()
        self.assertEqual(self.called, 1)

    def test_realGetAllThemes(self):
        """
        C{_realGetAllThemes} should collect themes from available offerings.
        """
        tc = webtheme.ThemeCache()
        ths = tc.getAllThemes()
        self.assertEqual([theme.name for theme in ths],
                         ['foo', 'boz', 'baz'])


    def test_clearThemeCache(self):
        """
        C{emptyCache} should invalidate the cache contents for both types.
        """
        tc = webtheme.ThemeCache()
        s = Store()
        tc.getAllThemes()
        tc.getInstalledThemes(s)
        tc.emptyCache()
        self.assertEqual(tc._getAllThemesCache, None)
        self.assertEqual(len(tc._getInstalledThemesCache), 0)



import sys

from nevow import tags
from nevow.livetrial.testcase import TestCase

from axiom.store import Store
from axiom.dependency import installOn

from xmantissa import people, ixmantissa
from xmantissa.liveform import FORM_INPUT
from xmantissa.webtheme import getLoader
from xmantissa.webapp import PrivateApplication

class AddPersonTestBase(people.AddPersonFragment):
    jsClass = None

    def __init__(self):
        self.store = Store()
        organizer = people.Organizer(store=self.store)
        installOn(organizer, self.store)
        people.AddPersonFragment.__init__(self, organizer)


    def getWidgetDocument(self):
        return tags.invisible(render=tags.directive('addPersonForm'))


    def mangleDefaults(self, params):
        """
        Called before rendering the form to give tests an opportunity to
        modify the defaults for the parameters being used.

        @type params: L{list} of liveform parameters
        @param params: The parameters which will be used by the liveform
            being rendered.
        """


    def checkResult(self, positional, keyword):
        """
        Verify that the given arguments are the ones which were expected by
        the form submission.  Override this in a subclass.

        @type positional: L{tuple}
        @param positional: The positional arguments submitted by the form.

        @type keyword: L{dict}
        @param keyword: The keyword arguments submitted by the form.
        """
        raise NotImplementedError()


    def addPerson(self, *a, **k):
        """
        Override form handler to just check the arguments given without
        trying to modify any database state.
        """
        self.checkResult(a, k)


    def render_addPersonForm(self, ctx, data):
        liveform = super(AddPersonTestBase, self).render_addPersonForm(ctx, data)

        # XXX This is a pretty terrible hack.  The client-side of these tests
        # just submit the form.  In order for the assertions to succeed, that
        # means the form needs to be rendered with some values in it already.
        # There's no actual API for putting values into the form here, though.
        # So instead, we'll grovel over all the parameters and try to change
        # them to reflect what we want.  Since this relies on there being no
        # conflictingly named parameters anywhere in the form and since it
        # relies on the parameters being traversable in order to find them all,
        # this is rather fragile.  The tests should most likely just put values
        # in on the client or something along those lines (it's not really
        # clear what the intent of these tests are, anyway, so it's not clear
        # what alternate approach would satisfy that intent).
        params = []
        remaining = liveform.parameters[:]
        while remaining:
            p = remaining.pop()
            if p.type == FORM_INPUT:
                remaining.extend(p.coercer.parameters)
            else:
                params.append((p.name, p))
        self.mangleDefaults(dict(params))
        return liveform



class OnlyNick(AddPersonTestBase, TestCase):
    jsClass = u'Mantissa.Test.OnlyNick'

    def mangleDefaults(self, params):
        """
        Set the nickname in the form to a particular value for
        L{checkResult} to verify when the form is submitted.
        """
        params['nickname'].default = u'everybody'


    def checkResult(self, positional, keyword):
        """
        There should be no positional arguments but there should be keyword
        arguments for each of the two attributes of L{Person} and three more
        for the basic contact items.  Only the nickname should have a value.
        """
        self.assertEqual(positional, ())
        self.assertEqual(
            keyword,
            {'nickname': u'everybody',
             'vip': False,
             'xmantissa.people.PostalContactType': [{'address': u''}],
             'xmantissa.people.EmailContactType': [{'email': u''}]})



class NickNameAndEmailAddress(AddPersonTestBase, TestCase):
    jsClass = u'Mantissa.Test.NickNameAndEmailAddress'

    def mangleDefaults(self, params):
        """
        Set the nickname and email address to values which L{checkResult}
        can verify.
        """
        params['nickname'].default = u'NICK!!!'
        params['xmantissa.people.EmailContactType'].parameters[0].default = u'a@b.c'


    def checkResult(self, positional, keyword):
        """
        Verify that the nickname and email address set in L{mangleDefaults}
        are submitted.
        """
        self.assertEqual(positional, ())
        self.assertEqual(
            keyword,
            {'nickname': u'NICK!!!',
             'vip': False,
             'xmantissa.people.PostalContactType': [{'address': u''}],
             'xmantissa.people.EmailContactType': [{'email': u'a@b.c'}]})



# -*- test-case-name: xmantissa.test.test_recordattr -*-

"""
Utility support for attributes on items which compose multiple Axiom attributes
into a single epsilon.structlike.record attribute.  This can be handy when
composing a simple, common set of columns that several tables share into a
recognizable object that is not an item itself.  For example, the pair of
'localpart', 'domain' into a user object, or the triple of 'realname',
'nickname', 'hostmask', 'network' into an IRC nickname.  This functionality is
currently used to make L{sharing.Identifier} objects.

This is a handy utility that should really be moved to L{axiom.attributes} and
made public as soon as a few conditions are met:

    * L{WithRecordAttributes} needs to be integrated into L{Item}, or
      otherwise made obsolete such that normal item instantiation works and
      users don't need to call a bogus classmethod.

    * L{RecordAttribute} needs to implement the full set of comparison
      operators required by the informal axiom constraint language (__gt__,
      __lt__, __ge__, __le__, probably some other stuff).  It would also be
      great if that informal language got documented somewhere.

"""

from axiom.attributes import AND

class RecordAttribute(object):
    """
    A descriptor which maps a group of axiom attributes into a single attribute
    which returns a record composing them all.

    Use this within an Item class definition, like so::

        class Address(record('localpart domain')):
            'An email address.'

        class Email(Item, WithRecordAttributes):
            senderLocalpart = text()
            senderDomain = text()
            receipientLocalpart = text()
            recipientDomain = text()
            body = text()

            sender = RecordAttribute(Address, senderLocalpart, senderDomain)
            recipient = RecordAttribute(Address, recipientLocalpart,
                                        recipientDomain)
        # ...

        myEmail = Email._recordCreate(sender=Address(localpart=u'hello',
                                                     domain=u'example.com'),
                                      recipient=Address(localpart=u'goodbye',
                                                        domain=u'example.com'))
        print myEmail.sender.localpart

    Note: the ugly _recordCreate method is required to create items which use
    this feature due to some problems with Axiom's initialization order.  See
    L{WithRecordAttributes} for details.
    """

    def __init__(self, recordType, attrs):
        """
        Create a L{RecordAttribute} for a certain record type and set of Axiom
        attributes.

        @param recordType: the result, or a subclass of the result, of
        L{axiom.structlike.record}.

        @param attrs: a tuple of L{axiom.attributes.SQLAttribute} instances
        that were defined as part of the schema on the same item type.
        """
        self.recordType = recordType
        self.attrs = attrs


    def __get__(self, oself, type=None):
        """
        Retrieve this compound attribute from the given item.

        @param oself: an L{axiom.item.Item} instance, of a type which has this
        L{RecordAttribute}'s L{attrs} defined in its schema.
        """
        if oself is None:
            return self
        constructData = {}
        for n, attr in zip(self.recordType.__names__, self.attrs):
            constructData[n] = attr.__get__(oself, type)
        return self.recordType(**constructData)


    def _decompose(self, value):
        """
        Decompose an instance of our record type into a dictionary mapping
        attribute names to values.

        @param value: an instance of self.recordType

        @return: L{dict} containing the keys declared on L{record}.
        """
        data = {}
        for n, attr in zip(self.recordType.__names__, self.attrs):
            data[attr.attrname] = getattr(value, n)
        return data


    def __set__(self, oself, value):
        """
        Set each component attribute of this L{RecordAttribute} in turn.

        @param oself: an instance of the type where this attribute is defined.

        @param value: an instance of self.recordType whose values should be
        used.
        """
        for n, attr in zip(self.recordType.__names__, self.attrs):
            attr.__set__(oself, getattr(value, n))


    def __eq__(self, other):
        """
        @return: a comparison object resulting in all of the component
        attributes of this attribute being equal to all of the attribute values
        on the other object.

        @rtype: L{IComparison}
        """
        return AND(*[attr == getattr(other, name)
                     for attr, name
                     in zip(self.attrs, self.recordType.__names__)])


    def __ne__(self, other):
        """
        @return: a comparison object resulting in all of the component
        attributes of this attribute being unequal to all of the attribute
        values on the other object.

        @rtype: L{IComparison}
        """
        return AND(*[attr != getattr(other, name)
                     for attr, name
                     in zip(self.attrs, self.recordType.__names__)])



class WithRecordAttributes(object):
    """
    Axiom has an unfortunate behavior, which is a rather deep-seated bug in the
    way Item objects are initialized.  Default parameters are processed before
    the attributes in the constructor's dictionary are actually set.  In other
    words, if you have a custom descriptor like L{RecordAttribute}, it can't be
    passed in the constructor; if the public way to fill in a required
    attribute's value is via such an API, it becomes impossible to properly
    construct an object.

    This mixin implements a temporary workaround, by adding a classmethod for
    creating instances of classes that use L{RecordAttribute} by explicitly
    decomposing the structured record instances into their constitutent values
    before actually passing them on to L{Item.__init__}.

    This workaround needs to be promoted to a proper resolution before this can
    be a public API; users should be able to create their own descriptors that
    modify underlying database state and have them behave in the expected way
    during item creation.
    """

    def create(cls, **kw):
        """
        Create an instance of this class, first cleaning up the keyword
        arguments so they will fill in any required values.

        @return: an instance of C{cls}
        """
        for k, v in kw.items():
            attr = getattr(cls, k, None)
            if isinstance(attr, RecordAttribute):
                kw.pop(k)
                kw.update(attr._decompose(v))
        return cls(**kw)
    create = classmethod(create)




from epsilon import hotfix
hotfix.require('twisted', 'trial_assertwarns')

import sha

from zope.interface import implements
from zope.interface.verify import verifyObject

try:
    from cssutils import CSSParser
    CSSParser
except ImportError:
    CSSParser = None

from twisted.python.components import registerAdapter
from twisted.internet.address import IPv4Address
from twisted.trial.unittest import TestCase
from twisted.trial import util
from twisted.python.filepath import FilePath

from nevow import tags
from nevow.flat import flatten
from nevow.context import WebContext
from nevow.testutil import FakeRequest
from nevow.url import URL
from nevow.inevow import IResource, IRequest
from nevow.rend import WovenContext, NotFound
from nevow.athena import LivePage, LiveElement, AthenaModule, jsDeps
from nevow.guard import LOGIN_AVATAR
from nevow.loaders import stan

from axiom import userbase
from axiom.userbase import LoginSystem
from axiom.store import Store
from axiom.dependency import installOn
from axiom.plugins.mantissacmd import Mantissa

from xmantissa.ixmantissa import (
    IProtocolFactoryFactory, ISiteURLGenerator, ISiteRootPlugin, IWebViewer,
    ISessionlessSiteRootPlugin)
from xmantissa.port import TCPPort, SSLPort
from xmantissa import website, publicweb
from xmantissa._webutil import VirtualHostWrapper

from xmantissa.publicweb import LoginPage
from xmantissa.offering import installOffering
from xmantissa.plugins.baseoff import baseOffering
from xmantissa.cachejs import theHashModuleProvider
from xmantissa.website import WebSite
from xmantissa.webapp import PrivateApplication
from xmantissa.websharing import SharingIndex

from xmantissa.website import MantissaLivePage, APIKey, PrefixURLMixin
from xmantissa.web import SecuringWrapper, _SecureWrapper, StaticContent, UnguardedWrapper, SiteConfiguration


maybeEncryptedRootWarningMessage = (
    "Use ISiteURLGenerator.rootURL instead of WebSite.maybeEncryptedRoot")
maybeEncryptedRootSuppression = util.suppress(
    message=maybeEncryptedRootWarningMessage,
    category=DeprecationWarning)



class SiteConfigurationTests(TestCase):
    """
    L{xmantissa.web.Site} defines how to create an HTTP server.
    """
    def setUp(self):
        self.domain = u"example.com"
        self.store = Store()
        self.site = SiteConfiguration(store=self.store, hostname=self.domain)


    def test_interfaces(self):
        """
        L{SiteConfiguration} implements L{IProtocolFactoryFactory} and
        L{ISiteURLGenerator}.
        """
        self.assertTrue(verifyObject(ISiteURLGenerator, self.site))
        self.assertTrue(verifyObject(IProtocolFactoryFactory, self.site))


    def _baseTest(self, portType, scheme, portNumber, method):
        portType(store=self.store, portNumber=portNumber, factory=self.site)
        self.assertEquals(
            getattr(self.site, method)(), URL(scheme, self.domain))


    def test_cleartextRoot(self):
        """
        L{SiteConfiguration.cleartextRoot} method returns the proper URL for
        HTTP communication with this site.
        """
        self._baseTest(TCPPort, 'http', 80, 'cleartextRoot')


    def test_encryptedRoot(self):
        """
        L{SiteConfiguration.encryptedRoot} method returns the proper URL for
        HTTPS communication with this site.
        """
        self._baseTest(SSLPort, 'https', 443, 'encryptedRoot')


    def _nonstandardPortTest(self, portType, scheme, portNumber, method):
        portType(store=self.store, portNumber=portNumber, factory=self.site)
        self.assertEquals(
            getattr(self.site, method)(),
            URL(scheme, '%s:%s' % (self.domain, portNumber)))


    def test_cleartextRootNonstandardPort(self):
        """
        L{SiteConfiguration.cleartextRoot} method returns the proper URL for
        HTTP communication with this site even if the server is listening on a
        non-standard port number.
        """
        self._nonstandardPortTest(TCPPort, 'http', 8000, 'cleartextRoot')


    def test_encryptedRootNonstandardPort(self):
        """
        L{SiteConfiguration.encryptedRoot} method returns the proper URL for
        HTTPS communication with this site even if the server is listening on a
        non-standard port number.
        """
        self._nonstandardPortTest(SSLPort, 'https', 8443, 'encryptedRoot')


    def _unavailableTest(self, method):
        self.assertEquals(getattr(self.site, method)(), None)


    def test_cleartextRootUnavailable(self):
        """
        L{SiteConfiguration.cleartextRoot} method returns None if there is no
        HTTP server listening.
        """
        self._unavailableTest('cleartextRoot')


    def test_encryptedRootUnavailable(self):
        """
        L{SiteConfiguration.encryptedRoot} method returns None if there is no
        HTTPS server listening.
        """
        self._unavailableTest('encryptedRoot')


    def _hostOverrideTest(self, portType, scheme, portNumber, method):
        portType(store=self.store, portNumber=portNumber, factory=self.site)
        self.assertEquals(
            getattr(self.site, method)(u'example.net'),
            URL(scheme, 'example.net'))


    def test_cleartextRootHostOverride(self):
        """
        A hostname passed to L{SiteConfiguration.cleartextRoot} overrides the
        configured hostname in the result.
        """
        self._hostOverrideTest(TCPPort, 'http', 80, 'cleartextRoot')


    def test_encryptedRootHostOverride(self):
        """
        A hostname passed to L{SiteConfiguration.encryptedRoot} overrides the
        configured hostname in the result.
        """
        self._hostOverrideTest(SSLPort, 'https', 443, 'encryptedRoot')


    def _portZero(self, portType, scheme, method):
        randomPort = 7777

        class FakePort(object):
            def getHost(self):
                return IPv4Address('TCP', u'example.com', randomPort)

        port = portType(store=self.store, portNumber=0, factory=self.site)
        port.listeningPort = FakePort()
        self.assertEquals(
            getattr(self.site, method)(),
            URL(scheme, '%s:%s' % (self.domain, randomPort)))


    def test_cleartextRootPortZero(self):
        """
        When the C{portNumber} of a started L{TCPPort} which refers to the
        C{SiteConfiguration} is C{0}, L{SiteConfiguration.cleartextRoot}
        returns an URL with the port number which was actually bound in the
        netloc.
        """
        self._portZero(TCPPort, 'http', 'cleartextRoot')


    def test_encryptedRootPortZero(self):
        """
        When the C{portNumber} of a started L{SSLPort} which refers to the
        C{SiteConfiguration} is C{0}, L{SiteConfiguration.encryptedRoot}
        returns an URL with the port number which was actually bound in the
        netloc.
        """
        self._portZero(SSLPort, 'https', 'encryptedRoot')


    def _portZeroDisconnected(self, portType, method):
        portType(store=self.store, portNumber=0, factory=self.site)
        self.assertEquals(None, getattr(self.site, method)())


    def test_cleartextRootPortZeroDisconnected(self):
        """
        When the C{portNumber} of an unstarted L{TCPPort} which refers to the
        C{SiteConfiguration} is C{0}, L{SiteConfiguration.cleartextRoot}
        returns C{None}.
        """
        self._portZeroDisconnected(TCPPort, 'cleartextRoot')


    def test_encryptedRootPortZeroDisconnected(self):
        """
        When the C{portNumber} of an unstarted L{SSLPort} which refers to the
        C{SiteConfiguration} is C{0}, L{SiteConfiguration.encryptedRoot}
        returns C{None}.
        """
        self._portZeroDisconnected(SSLPort, 'encryptedRoot')


    def test_rootURL(self):
        """
        L{SiteConfiguration.rootURL} returns C{/} for a request made onto the
        hostname with which the L{SiteConfiguration} is configured.
        """
        request = FakeRequest(headers={
            'host': self.domain.encode('ascii')})
        self.assertEqual(self.site.rootURL(request), URL('', ''))


    def test_rootURLWithoutHost(self):
        """
        L{SiteConfiguration.rootURL} returns C{/} for a request made without a
        I{Host} header.
        """
        request = FakeRequest()
        self.assertEqual(self.site.rootURL(request), URL('', ''))


    def test_rootURLWWWSubdomain(self):
        """
        L{SiteConfiguration.rootURL} returns C{/} for a request made onto the
        I{www} subdomain of the hostname of the L{SiteConfiguration}.
        """
        request = FakeRequest(headers={
            'host': 'www.' + self.domain.encode('ascii')})
        self.assertEqual(self.site.rootURL(request), URL('', ''))


    def _differentHostnameTest(self, portType, portNumber, isSecure, scheme):
        request = FakeRequest(isSecure=isSecure, headers={
            'host': 'alice.' + self.domain.encode('ascii')})
        portType(store=self.store, factory=self.site, portNumber=portNumber)
        self.assertEqual(self.site.rootURL(request), URL(scheme, self.domain))


    def test_cleartextRootURLDifferentHostname(self):
        """
        L{SiteConfiguration.rootURL} returns an absolute URL with the HTTP
        scheme and its hostname as the netloc and with a path of C{/} for a
        request made over HTTP onto a hostname different from the hostname of the
        L{SiteConfiguration}.

        """
        self._differentHostnameTest(TCPPort, 80, False, 'http')


    def test_encryptedRootURLDifferentHostname(self):
        """
        L{SiteConfiguration.rootURL} returns an absolute URL with its hostname
        as the netloc and with a path of C{/} for a request made over HTTPS onto a
        hostname different from the hostname of the L{SiteConfiguration}.
        """
        self._differentHostnameTest(SSLPort, 443, True, 'https')


    def _differentHostnameNonstandardPort(self, portType, isSecure, scheme):
        portNumber = 12345
        request = FakeRequest(isSecure=isSecure, headers={
            'host': 'alice.' + self.domain.encode('ascii')})
        portType(store=self.store, factory=self.site, portNumber=portNumber)
        self.assertEqual(
            self.site.rootURL(request),
            URL(scheme, '%s:%s' % (self.domain.encode('ascii'), portNumber)))


    def test_cleartextRootURLDifferentHostnameNonstandardPort(self):
        """
        L{SiteConfiguration.rootURL} returns an absolute URL with an HTTP
        scheme and an explicit port number in the netloc for a request made
        over HTTP onto a hostname different from the hostname of the
        L{SiteConfiguration} if the L{SiteConfiguration} has an HTTP server on
        a non-standard port.
        """
        self._differentHostnameNonstandardPort(TCPPort, False, 'http')


    def test_encryptedRootURLDifferentHostnameNonstandardPort(self):
        """
        L{SiteConfiguration.rootURL} returns an absolute URL with an HTTPS
        scheme and an explicit port number in the netloc for a request made
        over HTTPS onto a hostname different from the hostname of the
        L{SiteConfiguration} if the L{SiteConfiguration} has an HTTPS server on
        a non-standard port.
        """
        self._differentHostnameNonstandardPort(SSLPort, True, 'https')


    def test_rootURLNonstandardRequestPort(self):
        """
        L{SiteConfiguration.rootURL} returns C{/} for a request made onto a
        non-standard port which is one on which the L{SiteConfiguration} is
        configured to listen.
        """
        request = FakeRequest(headers={
            'host': '%s:%s' % (self.domain.encode('ascii'), 54321)})
        TCPPort(store=self.store, factory=self.site, portNumber=54321)
        self.assertEqual(self.site.rootURL(request), URL('', ''))



class StylesheetRewritingRequestWrapperTests(TestCase):
    """
    Tests for L{StylesheetRewritingRequestWrapper}.
    """
    def test_replaceMantissa(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} changes URLs of the form
        I{/Mantissa/foo} to I{<rootURL>/static/mantissa-base/foo}.
        """
        request = object()
        roots = {request: URL.fromString('/bar/')}
        wrapper = website.StylesheetRewritingRequestWrapper(request, [], roots.get)
        self.assertEqual(
            wrapper._replace('/Mantissa/foo.png'),
            '/bar/static/mantissa-base/foo.png')


    def test_replaceOtherOffering(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} changes URLs of the form
        I{/Something/foo} to I{<rootURL>/static/Something/foo} if C{Something}
        gives the name of an installed offering with a static content path.
        """
        request = object()
        roots = {request: URL.fromString('/bar/')}
        wrapper = website.StylesheetRewritingRequestWrapper(request, ['OfferingName'], roots.get)
        self.assertEqual(
            wrapper._replace('/OfferingName/foo.png'),
            '/bar/static/OfferingName/foo.png')


    def test_nonOfferingOnlyGivenPrefix(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} only changes URLs of the
        form I{/Something/foo} so they are beneath the root URL if C{Something}
        does not give the name of an installed offering.
        """
        request = object()
        roots = {request: URL.fromString('/bar/')}
        wrapper = website.StylesheetRewritingRequestWrapper(
            request, ['Foo'], roots.get)
        self.assertEqual(
            wrapper._replace('/OfferingName/foo.png'),
            '/bar/OfferingName/foo.png')


    def test_shortURL(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} changes URLs with only
        one segment so they are beneath the root URL.
        """
        request = object()
        roots = {request: URL.fromString('/bar/')}
        wrapper = website.StylesheetRewritingRequestWrapper(
            request, [], roots.get)
        self.assertEqual(
            wrapper._replace('/'),
            '/bar/')


    def test_absoluteURL(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} does not change absolute
        URLs.
        """
        wrapper = website.StylesheetRewritingRequestWrapper(object(), [], None)
        self.assertEqual(
            wrapper._replace('http://example.com/foo'),
            'http://example.com/foo')


    def test_relativeUnmodified(self):
        """
        L{StylesheetRewritingRequestWrapper._replace} does not change URLs with
        relative paths.
        """
        wrapper = website.StylesheetRewritingRequestWrapper(object(), [], None)
        self.assertEqual(wrapper._replace('relative/path'), 'relative/path')


    def test_finish(self):
        """
        L{StylesheetRewritingRequestWrapper.finish} causes all written bytes to
        be translated with C{_replace} written to the wrapped request.
        """
        stylesheetFormat = """
            .foo {
                background-image: url(%s)
            }
        """
        originalStylesheet = stylesheetFormat % ("/Foo/bar",)
        expectedStylesheet = stylesheetFormat % ("/bar/Foo/bar",)

        request = FakeRequest()
        roots = {request: URL.fromString('/bar/')}
        wrapper = website.StylesheetRewritingRequestWrapper(
            request, [], roots.get)
        wrapper.write(originalStylesheet)
        wrapper.finish()
        # Parse and serialize both versions to normalize whitespace so we can
        # make a comparison.
        parser = CSSParser()
        self.assertEqual(
            parser.parseString(request.accumulator).cssText,
            parser.parseString(expectedStylesheet).cssText)
    if CSSParser is None:
        test_finish.skip = "Stylesheet rewriting test requires cssutils package."



class LoginPageTests(TestCase):
    """
    Tests for functionality related to login.
    """
    domain = u"example.com"

    def setUp(self):
        """
        Create a L{Store}, L{WebSite} and necessary request-related objects to
        test L{LoginPage}.
        """
        self.siteStore = Store(filesdir=self.mktemp())
        Mantissa().installSite(self.siteStore, self.domain, u"", False)
        self.site = self.siteStore.findUnique(SiteConfiguration)
        installOn(
            TCPPort(store=self.siteStore, factory=self.site, portNumber=80),
            self.siteStore)
        self.context = WebContext()
        self.request = FakeRequest()
        self.context.remember(self.request)


    def test_fromRequest(self):
        """
        L{LoginPage.fromRequest} should return a two-tuple of the class it is
        called on and an empty tuple.
        """
        request = FakeRequest(
            uri='/foo/bar/baz',
            currentSegments=['foo'],
            args={'quux': ['corge']})

        class StubLoginPage(LoginPage):
            def __init__(self, store, segments, arguments):
                self.store = store
                self.segments = segments
                self.arguments = arguments

        page = StubLoginPage.fromRequest(self.siteStore, request)
        self.assertTrue(isinstance(page, StubLoginPage))
        self.assertIdentical(page.store, self.siteStore)
        self.assertEqual(page.segments, ['foo', 'bar'])
        self.assertEqual(page.arguments, {'quux': ['corge']})


    def test_staticShellContent(self):
        """
        The L{IStaticShellContent} adapter for the C{store} argument to
        L{LoginPage.__init__} should become its C{staticContent} attribute.
        """
        originalInterface = publicweb.IStaticShellContent
        adaptions = []
        result = object()
        def stubInterface(object, default):
            adaptions.append((object, default))
            return result
        publicweb.IStaticShellContent = stubInterface
        try:
            page = LoginPage(self.siteStore)
        finally:
            publicweb.IStaticShellContent = originalInterface
        self.assertEqual(len(adaptions), 1)
        self.assertIdentical(adaptions[0][0], self.siteStore)
        self.assertIdentical(page.staticContent, result)


    def test_segments(self):
        """
        L{LoginPage.beforeRender} should fill the I{login-action} slot with an
        L{URL} which includes all the segments given to the L{LoginPage}.
        """
        segments = ('foo', 'bar')
        page = LoginPage(self.siteStore, segments)
        page.beforeRender(self.context)
        loginAction = self.context.locateSlotData('login-action')
        expectedLocation = URL.fromString('/')
        for segment in (LOGIN_AVATAR,) + segments:
            expectedLocation = expectedLocation.child(segment)
        self.assertEqual(loginAction, expectedLocation)


    def test_queryArguments(self):
        """
        L{LoginPage.beforeRender} should fill the I{login-action} slot with an
        L{URL} which includes all the query arguments given to the
        L{LoginPage}.
        """
        args = {'foo': ['bar']}
        page = LoginPage(self.siteStore, (), args)
        page.beforeRender(self.context)
        loginAction = self.context.locateSlotData('login-action')
        expectedLocation = URL.fromString('/')
        expectedLocation = expectedLocation.child(LOGIN_AVATAR)
        expectedLocation = expectedLocation.add('foo', 'bar')
        self.assertEqual(loginAction, expectedLocation)


    def test_locateChildPreservesSegments(self):
        """
        L{LoginPage.locateChild} should create a new L{LoginPage} with segments
        extracted from the traversal context.
        """
        segments = ('foo', 'bar')
        page = LoginPage(self.siteStore)
        child, remaining = page.locateChild(self.context, segments)
        self.assertTrue(isinstance(child, LoginPage))
        self.assertEqual(remaining, ())
        self.assertEqual(child.segments, segments)


    def test_locateChildPreservesQueryArguments(self):
        """
        L{LoginPage.locateChild} should create a new L{LoginPage} with query
        arguments extracted from the traversal context.
        """
        self.request.args = {'foo': ['bar']}
        page = LoginPage(self.siteStore)
        child, remaining = page.locateChild(self.context, None)
        self.assertTrue(isinstance(child, LoginPage))
        self.assertEqual(child.arguments, self.request.args)



class UnguardedWrapperTests(TestCase):
    """
    Tests for L{UnguardedWrapper}.
    """
    def setUp(self):
        """
        Set up a store with a valid offering to test against.
        """
        self.store = Store()
        installOffering(self.store, baseOffering, {})
        self.site = ISiteURLGenerator(self.store)


    def test_live(self):
        """
        L{UnguardedWrapper} has a I{live} child which returns a L{LivePage}
        instance.
        """
        request = FakeRequest(uri='/live/foo', currentSegments=[])
        wrapper = UnguardedWrapper(self.store, None)
        resource = wrapper.child_live(request)
        self.assertTrue(isinstance(resource, LivePage))


    def test_jsmodules(self):
        """
        L{UnguardedWrapper} has a I{__jsmodules__} child which returns a
        L{LivePage} instance.
        """
        request = FakeRequest(uri='/__jsmodule__/foo', currentSegments=[])
        wrapper = UnguardedWrapper(None, None)
        resource = wrapper.child___jsmodule__(request)

        # This is weak.  Identity of this object doesn't matter.  The caching
        # and jsmodule serving features are what matter. -exarkun
        self.assertIdentical(resource, theHashModuleProvider)


    def test_static(self):
        """
        L{UnguardedWrapper} has a I{static} child which returns a
        L{StaticContent} instance.
        """
        request = FakeRequest(uri='/static/extra', currentSegments=[])
        wrapper = UnguardedWrapper(self.store, None)
        resource = wrapper.child_static(request)
        self.assertTrue(isinstance(resource, StaticContent))
        self.assertEqual(
            resource.staticPaths,
            {baseOffering.name: baseOffering.staticContentPath})


    def test_sessionlessPlugin(self):
        """
        L{UnguardedWrapper.locateChild} looks up L{ISessionlessSiteRootPlugin}
        powerups on its store, and invokes their C{sessionlessProduceResource}
        methods to discover resources.
        """
        wrapper = UnguardedWrapper(self.store, None)
        req = FakeRequest()
        segments = ('foo', 'bar')
        calledWith = []
        result = object()
        class SiteRootPlugin(object):
            def sessionlessProduceResource(self, request, segments):
                calledWith.append((request, segments))
                return result
        self.store.inMemoryPowerUp(SiteRootPlugin(), ISessionlessSiteRootPlugin)

        resource = wrapper.locateChild(req, segments)
        self.assertEqual(calledWith, [(req, ("foo", "bar"))])


    def test_sessionlessLegacyPlugin(self):
        """
        L{UnguardedWrapper.locateChild} honors old-style
        L{ISessionlessSiteRootPlugin} providers that only implement a
        C{resourceFactory} method.
        """
        wrapper = UnguardedWrapper(self.store, None)
        req = FakeRequest()
        segments = ('foo', 'bar')
        calledWith = []
        result = object()
        class SiteRootPlugin(object):
            def resourceFactory(self, segments):
                calledWith.append(segments)
                return result
        self.store.inMemoryPowerUp(SiteRootPlugin(), ISessionlessSiteRootPlugin)

        resource = wrapper.locateChild(req, segments)
        self.assertEqual(calledWith, [segments])


    def test_confusedNewPlugin(self):
        """
        If L{UnguardedWrapper.locateChild} discovers a plugin that implements
        both C{sessionlessProduceResource} and C{resourceFactory}, it should
        prefer the new C{sessionlessProduceResource} method and return that
        resource.
        """
        wrapper = UnguardedWrapper(self.store, None)
        req = FakeRequest()
        test = self
        segments = ('foo', 'bar')
        result = object()
        calledWith = []
        class SiteRootPlugin(object):
            def resourceFactory(self, segments):
                test.fail("Don't call this.")
            def sessionlessProduceResource(self, request, segments):
                calledWith.append((request, segments))
                return result, segments[1:]
        self.store.inMemoryPowerUp(SiteRootPlugin(), ISessionlessSiteRootPlugin)

        resource, resultSegments = wrapper.locateChild(req, segments)
        self.assertEqual(calledWith, [(req, segments)])
        self.assertIdentical(resource, result)
        self.assertEqual(resultSegments, ('bar',))



class SiteTestsMixin(object):
    """
    Tests that apply to both subclasses of L{SiteRootMixin}, L{WebSite} and
    L{AnonymousSite}.

    @ivar store: a store containing something that inherits from L{SiteRootMixin}.

    @ivar resource: the result of adapting the given L{SiteRootMixin} item to
    L{IResource}.
    """

    def setUp(self):
        """
        Set up a site store with the base offering installed.
        """
        self.siteStore = Store()
        installOffering(self.siteStore, baseOffering, {})


    def addUser(self, username=u'nobody', domain=u'nowhere'):
        """
        Add a user to the site store, and install a L{PrivateApplication}.

        @return: the user store of the added user.
        """
        userStore = self.siteStore.findUnique(LoginSystem).addAccount(
            username, domain, u'asdf', internal=True).avatars.open()
        installOn(PrivateApplication(store=userStore), userStore)
        return userStore


    def test_crummyOldSiteRootPlugin(self):
        """
        L{SiteRootMixin.locateChild} queries for L{ISiteRootPlugin} providers
        and returns the result of their C{resourceFactory} method if it is not
        C{None}.
        """
        result = object()
        calledWith = []
        class SiteRootPlugin(object):
            def resourceFactory(self, segments):
                calledWith.append(segments)
                return result

        self.store.inMemoryPowerUp(SiteRootPlugin(), ISiteRootPlugin)
        self.assertIdentical(
            self.resource.locateChild(
                FakeRequest(headers={"host": "example.com"}),
                ("foo", "bar")),
            result)
        self.assertEqual(calledWith, [("foo", "bar")])


    def test_shinyNewSiteRootPlugin(self):
        """
        L{SiteRootMixin.locateChild} queries for L{ISiteRootPlugin} providers
        and returns the result of their C{produceResource} methods.
        """
        navthing = object()
        result = object()
        calledWith = []
        class GoodSiteRootPlugin(object):
            implements(ISiteRootPlugin)
            def produceResource(self, req, segs, webViewer):
                calledWith.append((req, segs, webViewer))
                return result

        self.store.inMemoryPowerUp(GoodSiteRootPlugin(), ISiteRootPlugin)
        self.store.inMemoryPowerUp(navthing, IWebViewer)
        req = FakeRequest(headers={
                'host': 'localhost'})
        self.resource.locateChild(req, ("foo", "bar"))
        self.assertEqual(calledWith, [(req, ("foo", "bar"), navthing)])


    def test_confusedNewPlugin(self):
        """
        When L{SiteRootMixin.locateChild} finds a L{ISiteRootPlugin} that
        implements both C{produceResource} and C{resourceFactory}, it should
        prefer the new-style C{produceResource} method.
        """
        navthing = object()
        result = object()
        calledWith = []
        test = self
        class ConfusedSiteRootPlugin(object):
            implements(ISiteRootPlugin)
            def produceResource(self, req, segs, webViewer):
                calledWith.append((req, segs, webViewer))
                return result
            def resourceFactory(self, segs):
                test.fail("resourceFactory called.")

        self.store.inMemoryPowerUp(ConfusedSiteRootPlugin(), ISiteRootPlugin)
        self.store.inMemoryPowerUp(navthing, IWebViewer)
        req = FakeRequest(headers={
                'host': 'localhost'})
        self.resource.locateChild(req, ("foo", "bar"))
        self.assertEqual(calledWith, [(req, ("foo", "bar"), navthing)])


    def test_virtualHosts(self):
        """
        When a virtual host of the form (x.y) is requested from a site root
        resource where 'y' is a known domain for that server, a L{SharingIndex}
        for the user identified as 'x@y' should be returned.
        """
        # Let's make sure we're looking at the correct store for our list of
        # domains first...
        self.assertIdentical(self.resource.siteStore,
                             self.siteStore)
        somebody = self.addUser(u'somebody', u'example.com')
        req = FakeRequest(headers={'host': 'somebody.example.com'})
        res, segs = self.resource.locateChild(req, ('',))
        self.assertIsInstance(res, SharingIndex) # :(
        self.assertIdentical(res.userStore, somebody)



class WebSiteTests(SiteTestsMixin, TestCase):
    """
    Isn't this class like four years late?
    """

    def setUp(self):
        """
        Set up a store with a valid offering to test against.
        """
        SiteTestsMixin.setUp(self)
        userStore = self.addUser()
        ws = userStore.findUnique(WebSite)
        self.store = ws.store
        self.resource = IResource(self.store)



class PrefixURLMixinTests(TestCase):
    """
    Tests for L{PrefixURLMixin}.
    """

    def test_createResource(self):
        """
        L{PrefixURLMixin} subclasses that implement C{createResource} have that
        method called upon resource lookup with no arguments.
        """
        target = object()
        class PrefixUser(PrefixURLMixin):
            prefixURL= u'foo'
            def createResource(self):
                return target

        p = PrefixUser()
        req = FakeRequest(headers={"host": "example.com"})
        rsrc, resultSegments = p.produceResource(req, ('foo', 'bar'), None)
        self.assertEqual(resultSegments, ('bar',))
        self.assertIdentical(rsrc, target)


    def test_createResourceWith(self):
        """
        L{PrefixURLMixin} subclasses that implement C{createResourceWith} have
        that method called upon resource lookup with the shell factory.
        """
        calledWith = []
        target = object()
        webViewer = object()
        class PrefixUser(PrefixURLMixin):
            prefixURL= u'foo'
            def createResourceWith(self, webViewer):
                calledWith.append(webViewer)
                return target

        p = PrefixUser()
        req = FakeRequest(headers={"host": "example.com"})
        rsrc, resultSegments = p.produceResource(req, ('foo', 'bar'),
                                                 webViewer)
        self.assertEqual(resultSegments, ('bar',))
        self.assertIdentical(rsrc, target)
        self.assertIdentical(webViewer, calledWith[0])


    def test_sessionlessResource(self):
        """
        L{PrefixURLMixin.sessionlessProduceResource} will produce the resource
        described by the L{createResource} for appropriate lists of segments.
        """
        target = object()
        class SessionlessPrefixThing(PrefixURLMixin):
            implements(ISessionlessSiteRootPlugin)
            prefixURL = u'foo'
            def createResource(self):
                return target

        p = SessionlessPrefixThing()
        verifyObject(ISessionlessSiteRootPlugin, p)
        req = FakeRequest(headers={"host": "example.com"})
        rsrc, segments = p.sessionlessProduceResource(req, ('foo', 'bar'))
        self.assertIdentical(rsrc, target)
        self.assertEqual(segments, ('bar',))

        result = p.sessionlessProduceResource(req, ('baz', 'boz'))
        self.assertIdentical(result, None)



class StubResource(object):
    """
    An L{IResource} implementation which behaves in a way which is useful for
    testing L{SecuringWrapper}.

    @ivar childResource: The object which will be returned as a child resource
        from C{locateChild}.
    @ivar childSegments: The object which will be returned as the
        unconsumed segments from C{locateChild}.
    @ivar renderContent: The object to return from C{renderHTTP}.
    @ivar renderedWithContext: The argument passed to C{renderHTTP}.
    """
    implements(IResource)

    def __init__(self, childResource, childSegments, renderContent):
        self.childResource = childResource
        self.childSegments = childSegments
        self.renderContent = renderContent


    def renderHTTP(self, context):
        self.renderedWithContext = context
        return self.renderContent


    def locateChild(self, context, segments):
        self.locatedWith = (context, segments)
        return self.childResource, self.childSegments



class NotResource(object):
    """
    A class which does not implement L{IResource}.
    """



class NotResourceAdapter(object):
    """
    An adapter from L{NotResource} to L{IResource} (not really - but close
    enough for the tests).
    """
    def __init__(self, notResource):
        self.notResource = notResource


registerAdapter(NotResourceAdapter, NotResource, IResource)



class SecuringWrapperTests(TestCase):
    """
    L{SecuringWrapper} makes sure that any resource which is eventually
    retrieved from a wrapped resource and then rendered is rendered over HTTPS
    if possible and desired.
    """
    def setUp(self):
        """
        Create a resource and a wrapper to test.
        """
        self.store = Store()
        self.urlGenerator = SiteConfiguration(store=self.store,
                                              hostname=u"example.com")
        self.child = StubResource(None, None, None)
        self.childSegments = ("baz", "quux")
        self.content = "some bytes perhaps"
        self.resource = StubResource(
            self.child, self.childSegments, self.content)
        self.wrapper = SecuringWrapper(self.urlGenerator, self.resource)


    def test_locateChildHTTPS(self):
        """
        L{SecuringWrapper.locateChild} returns the wrapped resource and the
        unmodified segments if it is called with a secure request.
        """
        segments = ("foo", "bar")
        request = FakeRequest(isSecure=True)
        newResource, newSegments = self.wrapper.locateChild(request, segments)
        self.assertIdentical(newResource, self.resource)
        self.assertEqual(newSegments, segments)


    def test_locateChildHTTP(self):
        """
        L{SecuringWrapper.locateChild} returns a L{_SecureWrapper} wrapped
        around its own wrapped resource along with the unmodified segments if
        it is called with an insecure request.
        """
        segments = ("foo", "bar")
        request = FakeRequest(isSecure=False)
        newResource, newSegments = self.wrapper.locateChild(request, segments)
        self.assertTrue(isinstance(newResource, _SecureWrapper))
        self.assertIdentical(newResource.urlGenerator, self.urlGenerator)
        self.assertIdentical(newResource.wrappedResource, self.resource)
        self.assertEqual(newSegments, segments)


    def test_renderHTTPNeedsSecure(self):
        """
        L{SecuringWrapper.renderHTTP} returns a L{URL} pointing at the same
        location as the request URI but with an https scheme if the wrapped
        resource has a C{needsSecure} attribute with a true value and the
        request is over http.
        """
        SSLPort(store=self.store, factory=self.urlGenerator, portNumber=443)
        request = FakeRequest(
            isSecure=False, uri='/bar/baz', currentSegments=['bar', 'baz'])
        self.resource.needsSecure = True
        result = self.wrapper.renderHTTP(request)
        self.assertEqual(
            result, URL('https', self.urlGenerator.hostname, ['bar', 'baz']))


    def test_renderHTTP(self):
        """
        L{SecuringWrapper.renderHTTP} returns the result of the wrapped
        resource's C{renderHTTP} method if the wrapped resource does not have a
        C{needsSecure} attribute with a true value.
        """
        request = FakeRequest(
            isSecure=False, uri='/bar/baz', currentSegments=['bar', 'baz'])
        result = self.wrapper.renderHTTP(request)
        self.assertIdentical(self.resource.renderedWithContext, request)
        self.assertEqual(result, self.content)


    def test_renderHTTPS(self):
        """
        L{SecuringWrapper.renderHTTP} returns the result of the wrapped
        resource's C{renderHTTP} method if it is called with a secure request.
        """
        request = FakeRequest(isSecure=True)
        result = self.wrapper.renderHTTP(request)
        self.assertIdentical(self.resource.renderedWithContext, request)
        self.assertEqual(result, self.content)


    def test_renderHTTPCannotSecure(self):
        """
        L{SecuringWrapper.renderHTTP} returns the result of the wrapped
        resource's C{renderHTTP} method if it is invoked over http but there is
        no https location available.
        """
        request = FakeRequest(isSecure=False)
        result = self.wrapper.renderHTTP(request)
        self.assertIdentical(self.resource.renderedWithContext, request)
        self.assertEqual(result, self.content)


    def test_childLocateChild(self):
        """
        L{_SecureWrapper.locateChild} returns a L{Deferred} which is called
        back with the result of the wrapped resource's C{locateChild} method
        wrapped in another L{_SecureWrapper}.
        """
        segments = ('foo', 'bar')
        request = FakeRequest()
        wrapper = _SecureWrapper(self.urlGenerator, self.resource)
        result = wrapper.locateChild(request, segments)
        def locatedChild((resource, segments)):
            self.assertTrue(isinstance(resource, _SecureWrapper))
            self.assertIdentical(resource.wrappedResource, self.child)
            self.assertEqual(segments, self.childSegments)
        result.addCallback(locatedChild)
        return result


    def test_notFound(self):
        """
        A L{_SecureWrapper.locateChild} lets L{NotFound} results from the
        wrapped resource pass through.
        """
        segments = ('foo', 'bar')
        request = FakeRequest()
        self.resource.childResource = None
        self.resource.childSegments = ()
        wrapper = _SecureWrapper(self.urlGenerator, self.resource)
        result = wrapper.locateChild(request, segments)
        def locatedChild(result):
            self.assertIdentical(result, NotFound)
        result.addCallback(locatedChild)
        return result


    def test_adaption(self):
        """
        A L{_SecureWrapper} constructed with an object which does not provide
        L{IResource} adapts it to L{IResource} and operates on the result.
        """
        notResource = NotResource()
        wrapper = _SecureWrapper(self.urlGenerator, notResource)
        self.assertTrue(isinstance(wrapper.wrappedResource, NotResourceAdapter))
        self.assertIdentical(wrapper.wrappedResource.notResource, notResource)



class AthenaResourcesTestCase(TestCase):
    """
    Test aspects of L{GenericNavigationAthenaPage}.
    """
    hostname = 'test-mantissa-live-page-mixin.example.com'

    def _preRender(self, resource):
        """
        Test helper which executes beforeRender on the given resource.

        This is used on live resources so that they don't start message queue
        timers.
        """
        ctx = WovenContext()
        req = FakeRequest(headers={'host': self.hostname})
        ctx.remember(req, IRequest)
        resource.beforeRender(ctx)


    def makeLivePage(self):
        """
        Create a MantissaLivePage instance for testing.
        """
        siteStore = Store(filesdir=self.mktemp())
        Mantissa().installSite(siteStore, self.hostname.decode('ascii'), u"", False)
        return MantissaLivePage(ISiteURLGenerator(siteStore))


    def test_transportRoot(self):
        """
        The transport root should always point at the '/live' transport root
        provided to avoid database interaction while invoking the transport.
        """
        livePage = self.makeLivePage()
        self.assertEquals(flatten(livePage.transportRoot), 'http://localhost/live')


    def test_debuggableMantissaLivePage(self):
        """
        L{MantissaLivePage.getJSModuleURL}'s depends on state from page
        rendering, but it should provide a helpful error message in the case
        where that state has not yet been set up.
        """
        livePage = self.makeLivePage()
        self.assertRaises(NotImplementedError, livePage.getJSModuleURL, 'Mantissa')


    def test_beforeRenderSetsModuleRoot(self):
        """
        L{MantissaLivePage.beforeRender} should set I{_moduleRoot} to the
        C{__jsmodule__} child of the URL returned by the I{rootURL}
        method of the L{WebSite} it wraps.
        """
        receivedRequests = []
        root = URL(netloc='example.com', pathsegs=['a', 'b'])
        class FakeWebSite(object):
            def rootURL(self, request):
                receivedRequests.append(request)
                return root
        request = FakeRequest()
        page = MantissaLivePage(FakeWebSite())
        page.beforeRender(request)
        self.assertEqual(receivedRequests, [request])
        self.assertEqual(page._moduleRoot, root.child('__jsmodule__'))


    def test_getJSModuleURL(self):
        """
        L{MantissaLivePage.getJSModuleURL} should return a child of its
        C{_moduleRoot} attribute of the form::

            _moduleRoot/<SHA1 digest of module contents>/Package.ModuleName
        """
        module = 'Mantissa'
        url = URL(scheme='https', netloc='example.com', pathsegs=['foo'])
        page = MantissaLivePage(None)
        page._moduleRoot = url
        jsDir = FilePath(__file__).parent().parent().child("js")
        modulePath = jsDir.child(module).child("__init__.js")
        moduleContents = modulePath.open().read()
        expect = sha.new(moduleContents).hexdigest()
        self.assertEqual(page.getJSModuleURL(module),
                         url.child(expect).child(module))


    def test_jsCaching(self):
        """
        Rendering a L{MantissaLivePage} causes each of its dependent modules to
        be loaded at most once.
        """
        _realdeps = AthenaModule.dependencies
        def dependencies(self):
            self.count = getattr(self, 'count', 0) + 1
            return _realdeps(self)
        self.patch(AthenaModule, 'dependencies', dependencies)

        module = jsDeps.getModuleForName('Mantissa.Test.Dummy.DummyWidget')
        module.count = 0

        root = URL(netloc='example.com', pathsegs=['a', 'b'])
        class FakeWebSite(object):
            def rootURL(self, request):
                return root

        def _renderPage():
            page = MantissaLivePage(FakeWebSite())
            element = LiveElement(stan(tags.span(render=tags.directive('liveElement'))))
            element.setFragmentParent(page)
            element.jsClass = u'Mantissa.Test.Dummy.DummyWidget'
            page.docFactory = stan([tags.span(render=tags.directive('liveglue')), element])

            ctx = WovenContext()
            req = FakeRequest(headers={'host': self.hostname})
            ctx.remember(req, IRequest)
            page.beforeRender(ctx)
            page.renderHTTP(ctx)
            page._messageDeliverer.close()

        # Make sure we have a fresh dependencies memo.
        self.patch(theHashModuleProvider, 'depsMemo', {})

        _renderPage()
        self.assertEqual(module.count, 1)
        _renderPage()
        self.assertEqual(module.count, 1)



class APIKeyTestCase(TestCase):
    """
    Tests for L{APIKey}.
    """
    def setUp(self):
        """
        Make a store.
        """
        self.store = Store()


    def test_getKeyForAPINone(self):
        """
        If there is no existing key for the named API, L{APIKey.getKeyForAPI}
        should return C{None}.
        """
        self.assertIdentical(
            APIKey.getKeyForAPI(self.store, u'this is an API name.'),
            None)


    def test_getKeyForAPIExisting(self):
        """
        If there is an existing key for the named API, L{APIKey.getKeyForAPI}
        should return it.
        """
        theAPIName = u'this is an API name.'
        existingAPIKey = APIKey(
            store=self.store,
            apiName=theAPIName,
            apiKey=u'this is an API key.')
        self.assertIdentical(
            existingAPIKey,
            APIKey.getKeyForAPI(self.store, theAPIName))


    def test_setKeyForAPINew(self):
        """
        If there is no existing key for the named API, L{APIKey.setKeyForAPI}
        should create a new L{APIKey} item.
        """
        theAPIKey = u'this is an API key.'
        theAPIName = u'this is an API name.'
        apiKey = APIKey.setKeyForAPI(
            self.store, theAPIName, theAPIKey)
        self.assertIdentical(apiKey, self.store.findUnique(APIKey))
        self.assertEqual(theAPIKey, apiKey.apiKey)
        self.assertEqual(theAPIName, apiKey.apiName)


    def test_setKeyForAPIExisting(self):
        """
        If there is an existing for the named API, L{APIKey.setKeyForAPI}
        should update its I{apiKey} attribute.
        """
        theAPIKey = u'this is an API key.'
        theAPIName = u'this is an API name.'
        existingAPIKey = APIKey(
            store=self.store, apiName=theAPIName, apiKey=theAPIKey)
        newAPIKey = u'this is a new API key'
        returnedAPIKey = APIKey.setKeyForAPI(
            self.store, theAPIName, newAPIKey)
        self.assertIdentical(existingAPIKey, returnedAPIKey)
        self.assertEqual(existingAPIKey.apiName, theAPIName)
        self.assertEqual(existingAPIKey.apiKey, newAPIKey)



class VirtualHostWrapperTests(TestCase):
    """
    Tests for L{VirtualHostWrapper}.
    """
    # XXX only integration tests cover L{VirtualHostWrapper.locateChild}.  That
    # is important functionality, it should be covered here.

    def test_nonSubdomain(self):
        """
        L{VirtualHostWrapper.subdomain} returns C{None} when passed a hostname
        which is not a subdomain of a domain of the site.
        """
        site = Store()
        wrapper = VirtualHostWrapper(site, None, None)
        self.assertIdentical(wrapper.subdomain("example.com"), None)


    def test_subdomain(self):
        """
        L{VirtualHostWrapper.subdomain} returns a two-tuple of a username and a
        domain name when passed a hostname which is a subdomain of a known
        domain.
        """
        site = Store()
        wrapper = VirtualHostWrapper(site, None, None)
        userbase.LoginMethod(
            store=site,
            account=site,
            protocol=u'*',
            internal=True,
            verified=True,
            localpart=u'alice',
            domain=u'example.com')
        self.assertEqual(
            wrapper.subdomain("bob.example.com"),
            ("bob", "example.com"))


    def test_wwwSubdomain(self):
        """
        L{VirtualHostWrapper.subdomain} returns C{None} when passed a hostname
        which is the I{www} subdomain of a domain of the site.
        """
        site = Store()
        wrapper = VirtualHostWrapper(site, None, None)
        userbase.LoginMethod(
            store=site,
            account=site,
            protocol=u'*',
            internal=True,
            verified=True,
            localpart=u'alice',
            domain=u'example.com')
        self.assertIdentical(wrapper.subdomain("www.example.com"), None)


    def test_subdomainWithPort(self):
        """
        L{VirtualHostWrapper.subdomain} handles hostnames with a port component
        as if they did not have a port component.
        """
        site = Store()
        wrapper = VirtualHostWrapper(site, None, None)
        userbase.LoginMethod(
            store=site,
            account=site,
            protocol=u'*',
            internal=True,
            verified=True,
            localpart=u'alice',
            domain=u'example.com')
        self.assertEqual(
            wrapper.subdomain("bob.example.com:8080"),
            ("bob", "example.com"))



# Copyright 2008 Divmod, Inc.  See LICENSE for details.

"""
Tests for L{xmantissa.stats}.
"""

from twisted.python import log, failure
from twisted.trial import unittest
from twisted.protocols.amp import ASK, COMMAND, Command, parseString

from axiom.iaxiom import IStatEvent

from xmantissa.stats import RemoteStatsCollectorFactory, RemoteStatsCollector
from xmantissa.test.test_ampserver import (
    BoxReceiverFactoryPowerupTestMixin, CollectingSender)



class RemoteStatsCollectorTest(BoxReceiverFactoryPowerupTestMixin, unittest.TestCase):
    """
    Tests for L{RemoteStatsCollectorFactory} and L{RemoteStatsCollector}.
    """
    factoryClass = RemoteStatsCollectorFactory
    protocolClass = RemoteStatsCollector

    def setUp(self):
        """
        Create and start a L{RemoteStatsCollector}.
        """
        self.receiver = RemoteStatsCollectorFactory().getBoxReceiver()
        self.sender = CollectingSender()
        self.receiver.startReceivingBoxes(self.sender)


    def tearDown(self):
        """
        Ensure the log observer added by L{setUp} is removed.
        """
        try:
            log.removeObserver(self.receiver._emit)
        except ValueError:
            # The test removed it already.
            pass


    def test_deliverStatEvents(self):
        """
        When a L{RemoteStatsCollector} is active, it sends AMP boxes
        to its client when L{IStatEvent}s are logged.
        """
        log.msg(interface=IStatEvent, foo="bar", baz=12, quux=u'\N{SNOWMAN}')
        self.assertEqual(len(self.sender.boxes), 1)
        stat = self.sender.boxes[0]
        self.assertNotIn(ASK, stat)
        self.assertEqual(stat[COMMAND], 'StatUpdate')
        # Skip testing the timestamp, another test which can control its value
        # will do that.
        data = set([
                (d['key'], d['value'])
                for d in parseString(stat['data'])
                if d['key'] != 'time'])
        self.assertEqual(
            data,
            set([('foo', 'bar'), ('baz', '12'),
                 ('quux', u'\N{SNOWMAN}'.encode('utf-8'))]))


    def test_ignoreOtherEvents(self):
        """
        L{RemoteStatsCollection} does not send any boxes for events which don't
        have L{IStatEvent} as the value for their C{'interface'} key.
        """
        log.msg(interface="test log event")
        self.assertEqual(self.sender.boxes, [])


    def test_deliveryStopsAfterDisconnect(self):
        """
        After L{RemoteStatsCollection.stopReceivingBoxes} is called, it no
        longer observes L{IStatEvent} log messages.
        """
        self.receiver.stopReceivingBoxes(
            failure.Failure(Exception("test exception")))
        log.msg(interface=IStatEvent, foo="bar")
        self.assertEqual(self.sender.boxes, [])


    def test_disconnectFailsOutgoing(self):
        """
        L{RemoteStatsCollection.stopReceivingBoxes} causes the L{Deferred}
        associated with any outstanding command to fail with the reason given.
        """
        class DummyCommand(Command):
            pass

        d = self.receiver.callRemote(DummyCommand)
        self.receiver.stopReceivingBoxes(RuntimeError("test exception"))
        return self.assertFailure(d, RuntimeError)


    def test_timestamp(self):
        """
        L{RemoteStatsCollector._emit} sends a I{StatUpdate} command with a
        timestamp taken from the log event passed to it.
        """
        self.receiver._emit({
                'time': 123456789.0,
                'interface': IStatEvent})
        self.assertEqual(len(self.sender.boxes), 1)
        timestamp = [
            d['value']
            for d in parseString(self.sender.boxes[0]['data'])
            if d['key'] == 'time']
        self.assertEqual(timestamp, ['123456789.0'])


    def test_athena(self):
        """
        L{RemoteStatsCollector._emit} sends a I{StatUpdate} command with Athena
        transport data for Athena message send and receive events.
        """
        self.receiver._emit({
                'athena_send_messages': True,
                'count': 17})
        self.receiver._emit({
                'athena_received_messages': True,
                'count': 12})
        self.assertEqual(len(self.sender.boxes), 2)
        send = set([(d['key'], d['value'])
                    for d in parseString(self.sender.boxes[0]['data'])])
        self.assertEqual(
            send,
            set([('count', '17'), ('athena_send_messages', 'True')]))
        received = set([(d['key'], d['value'])
                        for d in parseString(self.sender.boxes[1]['data'])])
        self.assertEqual(
            received,
            set([('count', '12'), ('athena_received_messages', 'True')]))



# Copyright 2008 Divmod, Inc. See LICENSE file for details

"""
This module tests the code listings used in the Mantissa lore documentation, in
the top-level 'doc/' directory.

See also L{nevow.test.test_howtolistings}.
"""

import sys

from zope.interface import implements

from zope.interface.verify import verifyObject

from twisted.python.filepath import FilePath

from twisted.trial.unittest import TestCase

from epsilon.structlike import record

from nevow.testutil import FakeRequest
from nevow.flat import flatten
from nevow.url import URL

from axiom.store import Store

from xmantissa.ixmantissa import ISiteRootPlugin, IMantissaSite, IWebViewer
from xmantissa.ixmantissa import IMessageReceiver
from xmantissa.publicweb import AnonymousSite
from axiom.plugins.mantissacmd import Mantissa


class StubShellPage(record('model')):
    """
    A stub implementation of the shell-page object that should be returned from
    wrapModel.
    """


class StubViewer(object):
    """
    A stub implementation of IWebViewer.
    """
    implements(IWebViewer)

    def wrapModel(self, thingy):
        """
        Wrap a model and return myself, with the model set.
        """
        self.shell = StubShellPage(thingy)
        return self.shell


class ExampleTestBase(object):
    """
    This mixin provides setUp and tearDown methods for adding a specific
    example to the path.

    See ticket #2713 for certain issues that this test has with operating in an
    installed environment.
    """

    def setUp(self):
        """
        Add the example dictated by C{self.examplePath} to L{sys.path}.
        """
        here = FilePath(__file__).parent().parent().parent().child('doc')
        for childName in self.examplePath:
            here = here.child(childName)
        sys.path.append(here.path)
        self.addCleanup(sys.path.remove, here.path)



class SiteRootDocumentationTest(ExampleTestBase, TestCase):
    """
    Tests for doc/siteroot.xhtml and friends.
    """

    examplePath = ['listings', 'siteroot']

    def setUp(self):
        """
        Do the example setup and import the module.
        """
        ExampleTestBase.setUp(self)
        import aboutpage
        import adminpage
        self.aboutpage = aboutpage
        self.adminpage = adminpage


    def test_powerItUp(self):
        """
        Powering up a store with an C{AboutPlugin} results in it being installed
        as an L{ISiteRootPlugin} powerup.
        """
        s = Store()
        ap = self.aboutpage.AboutPlugin(store=s)
        s.powerUp(ap)
        self.assertEquals([ap], list(s.powerupsFor(ISiteRootPlugin)))


    def test_interface(self):
        """
        C{AboutPlugin} implements L{ISiteRootPlugin}.
        """
        self.assertTrue(verifyObject(ISiteRootPlugin, self.aboutpage.AboutPlugin()))


    def test_produceAboutResource(self):
        """
        When C{AboutPlugin} is installed on a site store created by 'axiomatic
        mantissa', requests for 'about.php' will be responded to by a helpful
        message wrapped in a shell page.
        """
        s = Store(self.mktemp())
        s.powerUp(self.aboutpage.AboutPlugin(store=s))
        m = Mantissa()
        m.installSite(s, u"localhost", u"", False)
        root = IMantissaSite(s)
        viewer = StubViewer()
        result, segments = root.siteProduceResource(FakeRequest(),
                                                    tuple(['about.php']),
                                                    viewer)
        self.assertIdentical(result, viewer.shell)
        self.assertIsInstance(result.model, self.aboutpage.AboutText)


    def test_notOtherResources(self):
        """
        C{AboutPlugin} will only respond to about.php, not every page on the
        site.
        """
        s = Store(self.mktemp())
        s.powerUp(self.aboutpage.AboutPlugin(store=s))
        s.powerUp(AnonymousSite(store=s))
        root = IMantissaSite(s)
        viewer = StubViewer()
        result = root.siteProduceResource(FakeRequest(),
                                          tuple(['undefined']),
                                          viewer)
        self.assertIdentical(result, None)


    def test_rendering(self):
        """
        C{AboutText} should render a <div> with a string in it.
        """
        self.assertEquals("<div>Hello, world!</div>",
                          flatten(self.aboutpage.AboutText(u'Hello, world!')))


    def test_adminRedirect(self):
        """
        When the admin redirect is installed on a store, it should return an
        URL which should redirect to /private.
        """
        s = Store(self.mktemp())
        s.powerUp(self.adminpage.RedirectPlugin(store=s))
        m = Mantissa()
        m.installSite(s, u'localhost', u'', False)
        root = IMantissaSite(s)
        viewer = StubViewer()
        result, segments = root.siteProduceResource(FakeRequest(),
                                                    tuple(['admin.php']),
                                                    viewer)
        self.assertEquals(result, URL.fromString("http://localhost/private"))



class InterstoreMessagingDocumentationTests(ExampleTestBase, TestCase):
    """
    Tests for doc/interstore.xhtml and related files.
    """
    examplePath = ['listings', 'interstore']

    def setUp(self):
        """
        Import the interstore messaging example code.
        """
        ExampleTestBase.setUp(self)
        import cal
        self.cal = cal


    def test_powerUp(self):
        """
        L{Calendar} is an L{IMessageReceiver} powerup.
        """
        store = Store()
        calendar = self.cal.Calendar(store=store)
        self.assertTrue(verifyObject(IMessageReceiver, calendar))
        store.powerUp(calendar)
        self.assertEquals(
            list(store.powerupsFor(IMessageReceiver)), [calendar])




from axiom.test.historic import stubloader
from xmantissa.signup import UserInfoSignup

class FreeTicketSignupTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        fts = self.store.findUnique(UserInfoSignup)
        self.assertEqual(len(fts.product.types), 6)




"""
Test utilities providing coverage for implementations of Mantissa interfaces.
"""

from epsilon.descriptor import requiredAttribute

from xmantissa.webtheme import XHTMLDirectoryTheme


class XHTMLDirectoryThemeTestsMixin:
    """
    Mixin defining tests for themes based on L{XHTMLDirectoryTheme}.

    Mix this in to a L{twisted.trial.unittest.TestCase} and set C{theme} and
    C{offering}.

    @ivar theme: An instance of the theme to be tested.  Set this.
    @ivar offering: The offering which includes the theme.  Set this.
    """
    theme = requiredAttribute('theme')
    offering = requiredAttribute('offering')

    def test_stylesheet(self):
        """
        The C{stylesheetLocation} of the theme being tested identifies an
        existing file beneath the offering's static content path.
        """
        self.assertTrue(isinstance(self.theme, XHTMLDirectoryTheme))
        self.assertEqual(self.theme.stylesheetLocation[0], 'static')
        self.assertEqual(self.theme.stylesheetLocation[1], self.offering.name)
        path = self.offering.staticContentPath
        for segment in self.theme.stylesheetLocation[2:]:
            path = path.child(segment)
        self.assertTrue(path.exists(),
                        "Indicated stylesheet %r does not exist" % (path,))






# Copyright 2005 Divmod, Inc.  See LICENSE file for details

from axiom.test.historic import stubloader

from xmantissa.ixmantissa import IFulltextIndexer
from xmantissa.fulltext import PyLuceneIndexer

class PyLuceneIndexerTestCase(stubloader.StubbedTest):
    def testUpgrade(self):
        """
        The PyLuceneIndexer should be findable by its interface now.  It also
        should have been reset since it was most likely slightly corrupt, with
        respect to deleted documents.
        """
        index = IFulltextIndexer(self.store)
        self.failUnless(isinstance(index, PyLuceneIndexer))
        self.assertEqual(index.indexDirectory, 'foo.index')
        # we called reset(), and there are no indexed items
        self.assertEqual(index.indexCount, 0)
        self.assertEqual(index.installedOn, self.store)
